import json
from pathlib import Path

from davidia.main import _create_bare_app

app = _create_bare_app()

public_path = Path("storybook/src/openapi.json")
public_path.parent.mkdir(exist_ok=True)

with open(public_path, "w") as f:
    json.dump(app.openapi(), f)

print(f"OpenAPI schema saved to {public_path}")



import logging
from datetime import datetime
from enum import Enum
from pathlib import Path

import numpy as np
from PIL import Image as im
from pydantic import BaseModel

from ..models.messages import (
    AppendLineDataMessage,
    HeatmapData,
    ImageData,
    ImageDataMessage,
    LineData,
    LineParams,
    MultiLineDataMessage,
)
from ..models.parameters import Aspect, PlotConfig, ScaleType

logger = logging.getLogger("benchmarks")


class PlotType(str, Enum):
    """Class for plot type"""

    multiline = "multiline"
    add_data = "add_data"
    heatmap = "heatmap"
    image = "image"


class BenchmarkParams(BaseModel):
    """Parameters for benchmark"""

    plot_type: PlotType
    params: list[int | float]
    iterations: int = 1
    pause: float


def _timestamp():
    return datetime.now().strftime("%Y%m%d%H%M%S")


DATA_PATH = Path(__file__).parents[1] / "data"
BENCHMARK_HELP: dict[PlotType, str] = {}


ML_DEFAULT_PARAMS = [5, 10240]
BENCHMARK_HELP[PlotType.multiline] = f"""number of lines, number of initial points,
    {ML_DEFAULT_PARAMS}"""


def multiline(params: list[int | float]):
    f"""Generate messages for benchmarking the plotting of lines

    Parameters
    ----------
    params: [int, int] = {BENCHMARK_HELP[PlotType.multiline]}

    Returns
    -------
    Generator of messages
    """
    params.extend(ML_DEFAULT_PARAMS[len(params) :])
    lines, points = (int(p) for p in params)
    multilines = []
    logger.debug("Using %i points:", points)
    xi = np.arange(points, dtype=np.int32)
    for n in range(lines):
        offset = 1000 - n * 160
        x = xi - offset
        y = (x % 1000 + x % 100 + x % 10).astype(np.float64)
        multilines.append(
            LineData(key=_timestamp(), line_params=LineParams(), x=xi, y=y)
        )
    yield MultiLineDataMessage(ml_data=multilines)


AD_DEFAULT_PARAMS = [5, 10240, 512, 32]
BENCHMARK_HELP[PlotType.add_data] = f"""number of lines, number of initial points,
    number of additional points, number of batches,
    {AD_DEFAULT_PARAMS}"""


def add_data(params: list[int | float]):
    f"""Generate messages for benchmarking the plotting of extra data for lines

    Parameters
    ----------
    params: [int, int, int, int] = {BENCHMARK_HELP[PlotType.add_data]}

    Returns
    -------
    Generator of plot messages
    """
    params.extend(AD_DEFAULT_PARAMS[len(params) :])
    params = [int(p) for p in params]
    logger.debug("Using parameters: %s", params)
    total = list(params)
    lines, points, added, batches = (int(p) for p in params)
    yield from multiline(params[:2])

    total = points
    for _ in range(batches):
        multilines = []
        xi = np.arange(added, dtype=np.int32) + total
        for n in range(lines):
            offset = 1000 - n * 160
            x = xi - offset
            y = (x % 1000 + x % 100 + x % 10).astype(np.float64)
            multilines.append(
                LineData(key=_timestamp(), x=xi, y=y, line_params=LineParams())
            )

        total += added
        yield AppendLineDataMessage(al_data=multilines)


def get_image(cache: list, name: str, i: int):
    if len(cache) <= i:
        cache.append(np.asarray(im.open(DATA_PATH / name)))
    return cache[i]


IMAGES_CACHE = []


def image(params: list[int | float]):
    """Generate plot messages for benchmarking the plotting of an image

    Parameters
    ----------
    params : list[int|float]

    Returns
    -------
    Generator of plot messages

    """
    for i in range(10):
        values = get_image(IMAGES_CACHE, f"label-{i}.png", i)
        x_values = np.arange(values.shape[1])
        y_values = np.arange(values.shape[0])
        data = ImageData(values=values, aspect=Aspect.equal)
        plot_config = PlotConfig(
            x_label="x-axis",
            y_label="y-axis",
            x_values=x_values,
            y_values=y_values,
            title="image benchmarking plot",
        )
        yield ImageDataMessage(im_data=data, plot_config=plot_config)


HEATMAPS_CACHE = []


def heatmap(params: list[int | float]):
    """Generate plot messages for benchmarking the plotting of a heatmap

    Parameters
    ----------
    params : list[int|float]

    Returns
    -------
    Generator of plot messages

    """
    for i in range(10):
        values = get_image(HEATMAPS_CACHE, f"blobs-{i}.png", i)
        x_values = np.arange(values.shape[1])
        y_values = np.arange(values.shape[0])
        data = HeatmapData(
            values=values,
            domain=(0, 255),
            aspect=Aspect.auto,
            heatmap_scale=ScaleType.linear,
            colour_map="RdBu",
        )
        plot_config = PlotConfig(
            x_label="x-axis",
            y_label="y-axis",
            x_values=x_values,
            y_values=y_values,
            title="heatmap benchmarking plot",
        )
        yield ImageDataMessage(im_data=data, plot_config=plot_config)



"""
2D selections

All points are [x,y]
All angles in radians
"""

import logging
from math import atan2, cos, degrees, hypot, pi, radians, sin
from uuid import uuid4

from pydantic import BeforeValidator, Field, model_validator
from typing import Annotated, Type

from .parameters import DvDModel


def _make_id():
    return uuid4().hex[-8:]  # use last 8 characters only


Number = float | int


def _make_float(x: Number) -> float:
    return x if isinstance(x, float) else float(x)


Float = Annotated[float, BeforeValidator(_make_float)]


def _make_tuple_floats(v: tuple[Number, Number]) -> tuple[float, float]:
    if all(isinstance(i, float) for i in v):
        return v
    return (float(v[0]), float(v[1]))


FloatTuple = Annotated[tuple[float, float], BeforeValidator(_make_tuple_floats)]


class SelectionBase(DvDModel, validate_assignment=True):
    """Base class for representing any selection"""

    id: str = Field(default_factory=_make_id)
    name: str = ""
    colour: str | None = None
    alpha: float = 0.3
    fixed: bool = True
    start: FloatTuple


#    @property  # make read-only by omitting setter
#    def id(self):
#        return self.id


class AxialSelection(SelectionBase):
    """Class for representing the selection of part of an axis"""

    length: Float
    dimension: int

    def _end_get(self) -> tuple[float, float]:
        """Get end point"""
        dx = self.length if self.dimension == 0 else 0
        dy = self.length if self.dimension == 1 else 0
        return self.start[0] + dx, self.start[1] + dy

    def _end_set(self, end: tuple[float, float]) -> None:
        d = self.dimension
        s = self.start[d]
        e = end[d] if len(end) == 2 else end[0]
        if e < s:
            self.start = (e, self.start[1]) if d == 0 else (self.start[0], e)
            self.length = s - e
        else:
            self.length = e - s

    end = property(_end_get, _end_set)


class OrientableSelection(SelectionBase):
    """Base class for representing any orientable selection"""

    angle: Float = 0.0

    def __init__(self, degrees=None, **data):
        """Initialize angle in degrees"""
        if degrees is not None:
            if "angle" in data:
                raise ValueError("Both angle and degrees must not be specified")
            data["angle"] = radians(degrees)
        super().__init__(**data)

    @property
    def degrees(self) -> float:
        """Get angle in degrees"""
        return degrees(self.angle)

    def set_degrees(self, degrees: float):
        """Set angle in degrees"""
        self.angle = radians(degrees)


class LinearSelection(OrientableSelection):
    """Class for representing the selection of a line"""

    length: Float

    def __init__(self, degrees=None, **data):  # for pyright
        super().__init__(degrees=degrees, **data)

    @property
    def end(self) -> tuple[float, float]:
        """Get end point"""
        ll = self.length
        ang = self.alpha
        return cos(ang) * ll, sin(ang) * ll

    def set_end(self, x: float, y: float):
        """Set from end point"""
        dx = x - self.start[0]
        dy = y - self.start[1]
        self.angle = atan2(dy, dx)
        self.length = hypot(dx, dy)


class RectangularSelection(OrientableSelection):
    """Class for representing the selection of a rectangle"""

    lengths: FloatTuple

    def __init__(self, degrees=None, **data):  # for pyright
        super().__init__(degrees=degrees, **data)

    @property
    def end(self) -> tuple[float, float]:
        """Get end point"""
        a = self.alpha
        c = cos(a)
        s = sin(a)
        dx = self.lengths[0]
        dy = self.lengths[0]
        return self.start[0] + c * dx - s * dy, self.start[1] + s * dx + c * dy

    def set_end(self, x: float, y: float):
        """Set from end point (preserving angle of orientation)"""
        dx = x - self.start[0]
        dy = y - self.start[1]
        a = self.alpha
        c = cos(a)
        s = sin(a)
        self.lengths = c * dx + s * dy, -s * dx + c * dy


class PolygonalSelection(SelectionBase):
    """Class for representing the selection of a polygon"""

    points: list[tuple[float, float]]
    closed: bool

    def __init__(self, closed=True, **data):
        """Note start is ignored and duplicated from first point"""
        start = data.get("start", None)
        points = data.get("points", None)
        if start is not None:
            if len(start) != 2:
                raise ValueError("Start must have two values")
            if points is None or len(points) == 0:
                data["points"] = points = [start]
        else:
            if points is None or len(points) == 0:
                raise ValueError("At least one point must be specified")
            data["start"] = points[0]

        for p in points:
            if len(p) != 2:
                raise ValueError("Every point must have two values")
        super().__init__(closed=closed, **data)

    @model_validator(mode="after")
    def check_start(self) -> "PolygonalSelection":
        if len(self.points) == 0:
            raise ValueError("At least one point must be specified")
        pt = self.points[0]
        if self.start is not pt and (self.start[0] != pt[0] or self.start[1] != pt[1]):
            self.start = pt
            logging.warning("Overwriting start with first point")
        return self


class EllipticalSelection(OrientableSelection):
    """Class for representing the selection of an ellipse"""

    semi_axes: FloatTuple

    def __init__(self, degrees=None, **data):  # for pyright
        super().__init__(degrees=degrees, **data)


class CircularSelection(SelectionBase):
    """Class for representing the selection of a circle"""

    radius: Float


class CircularSectorialSelection(SelectionBase):
    """Class for representing the selection of a circular sector"""

    radii: FloatTuple
    angles: FloatTuple = (0.0, 2 * pi)

    def __init__(self, degrees=None, **data):
        """Initialize angles in degrees"""
        if degrees is not None:
            if "angles" in data:
                raise ValueError("Both angles and degrees must not be specified")
            data["angles"] = tuple(radians(d) for d in degrees)
        super().__init__(**data)

    @property
    def degrees(self) -> tuple[float, float]:
        """Get angle in degrees"""
        return (degrees(self.angles[0]), degrees(self.angles[1]))

    def set_degrees(self, degrees: tuple[Number, Number]):
        """Set angles in degrees"""
        self.angles = (radians(degrees[0]), radians(degrees[1]))


AnySelection = (
    AxialSelection
    | LinearSelection
    | RectangularSelection
    | PolygonalSelection
    | CircularSelection
    | EllipticalSelection
    | CircularSectorialSelection
    | SelectionBase
)


def as_selection(raw: dict | SelectionBase) -> AnySelection:
    if isinstance(raw, SelectionBase):
        return raw

    oc: Type[AnySelection]
    if "dimension" in raw:
        oc = AxialSelection
    elif "length" in raw:
        oc = LinearSelection
    elif "lengths" in raw:
        oc = RectangularSelection
    elif "points" in raw:
        oc = PolygonalSelection
    elif "semi_axes" in raw:
        oc = EllipticalSelection
    elif "radius" in raw:
        oc = CircularSelection
    elif "radii" in raw:
        oc = CircularSectorialSelection
    else:
        raise ValueError(f"Unknown selection that has {raw}")

    return oc.model_validate(raw)



import datetime
import logging
import time
from collections import defaultdict
from unittest.mock import AsyncMock, Mock

import before_after
import numpy as np
import pytest
from davidia.models.messages import (
    AppendLineDataMessage,
    ClearSelectionsMessage,
    DataMessage,
    DvDNDArray,
    LineData,
    LineParams,
    MsgType,
    MultiLineDataMessage,
    PlotMessage,
    SelectionsMessage,
    StatusType,
    TableData,
    TableDataMessage,
    UpdateSelectionsMessage,
)
from davidia.models.parameters import PlotConfig
from davidia.models.selections import LinearSelection, RectangularSelection
from davidia.server.fastapi_utils import ws_pack, ws_unpack
from davidia.server.plotserver import (
    PlotClient,
    PlotServer,
    PlotState,
    add_indices,
    convert_append_to_multi_line_data_message,
)
from davidia.server.processor import (
    check_line_names,
)

from .test_api import nppd_assert_equal


@pytest.fixture(autouse=True)
def no_error_logging(request, caplog):
    yield
    if "expect_caplog_errors" in request.keywords:
        return
    errors = [
        record
        for record in caplog.get_records("call")
        if record.levelno >= logging.ERROR
    ]
    assert not errors


def test_initialise_plotserver():
    ps = PlotServer()
    assert ps.client_status == StatusType.busy
    assert ps._clients == {}
    assert ps.plot_states == {}
    assert ps.client_total == 0
    assert not ps.clients_available()


@pytest.mark.asyncio  # @UndefinedVariable
async def test_send_points():
    ps = PlotServer()

    assert ps.client_status == StatusType.busy
    assert not ps.clients_available()
    assert ps.plot_states == {}

    plot_state_0 = ps.plot_states["plot_0"]
    x = np.array([i for i in range(50)])
    y = np.array([j % 10 for j in x])
    time_id = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    new_line = PlotMessage(
        plot_id="plot_0",
        type=MsgType.new_multiline_data,
        params=[
            {
                "key": time_id,
                "x": x,
                "y": y,
                "line_params": LineParams(colour="purple"),
            }
        ],
    )

    processed_line = ps.processor.process(new_line)
    line_as_dict = processed_line.model_dump(by_alias=True)

    msg = ws_pack(line_as_dict)
    assert msg is not None
    plot_state_0.current_data = line_as_dict  # pyright: ignore[reportGeneralTypeIssues]
    plot_state_0.new_data_message = msg
    assert not ps.clients_available()

    assert ps.client_status == StatusType.busy
    assert plot_state_0.current_data == line_as_dict
    assert plot_state_0.new_data_message == msg
    assert not ps.clients_available()

    await ps.send_next_message()

    assert ps.client_status == StatusType.busy
    assert plot_state_0.current_data == line_as_dict
    assert plot_state_0.new_data_message == msg
    assert not ps.clients_available()

    unpacked_msg = ws_unpack(msg)
    nppd_assert_equal(line_as_dict, unpacked_msg)


@pytest.mark.parametrize(
    "name,key,x,default_indices",
    [
        ("zero_x", "100", np.array([]), True),
        ("x_and_y", "101", np.array([0, 2]), False),
    ],
)
def test_line_data_initialization(name, key: str, x: list, default_indices: bool):
    a = LineData(
        key=key,
        line_params=LineParams(colour="blue", point_size=None),
        x=np.array(x),
        y=np.array([4, 2]),
    )
    assert a.default_indices == default_indices


def _array_equal(a: DvDNDArray | None, b: DvDNDArray | None):
    if a is None:
        return b is None
    if b is None:
        return False
    return np.array_equal(a, b)


def line_data_are_equal(a: LineData, b: LineData) -> bool:
    return (
        line_params_are_equal(a.line_params, b.line_params)
        and a.default_indices == b.default_indices
        and _array_equal(a.x, b.x)
        and np.array_equal(a.y, b.y)
    )


def line_params_are_equal(a: LineParams, b: LineParams) -> bool:
    return (
        a.colour == b.colour
        and a.line_on == b.line_on
        and a.point_size == b.point_size
        and a.glyph_type == b.glyph_type
        and a.name == b.name
    )


def assert_line_data_messages_are_equal(
    a: MultiLineDataMessage | AppendLineDataMessage,
    b: MultiLineDataMessage | AppendLineDataMessage,
):
    assert a.plot_config == b.plot_config
    if isinstance(a, MultiLineDataMessage) and isinstance(b, MultiLineDataMessage):
        assert len(a.ml_data) == len(b.ml_data)
        assert all([line_data_are_equal(c, d) for c, d in zip(a.ml_data, b.ml_data)])
    elif isinstance(a, AppendLineDataMessage) and isinstance(b, AppendLineDataMessage):
        assert len(a.al_data) == len(b.al_data)
        assert all([line_data_are_equal(c, d) for c, d in zip(a.al_data, b.al_data)])
    else:
        raise AssertionError(
            "a and b must both be either MultiLineDataMessage or"
            f" AppendLineDataMessage: {a}, {b}"
        )


def generate_test_data(
    key,
    x=True,
    default_indices=False,
    combined=False,
    high=False,
    name="",
    colour=None,
):
    if key == "100":
        if combined:
            x_data = [0, 1, 2, 3, 4, 5, 6]
            y_data = [10, 20, 30, 70, 60, 50, 40]
        else:
            x_data = [0, 1, 2]
            y_data = [10, 20, 30]

    elif key == "200":
        if combined:
            y_data = [5, 10, 15, 20, 50, 70, 90]
            x_data = [2, 3, 4, 7, 8, 9, 10] if high else [0, 1, 2, 3, 4, 5, 6]
        else:
            y_data = [5, 10, 15]
            x_data = [2, 3, 4] if high else [0, 1, 2]

    elif key == "300":
        y_data = [14, 12, 10, 8]
        x_data = [2, 3, 7, 9] if high else [0, 1, 2, 3]

    elif key == "010":
        x_data = [3, 4, 5, 6]
        y_data = [70, 60, 50, 40]

    elif key == "020":
        x_data = [7, 8, 9, 10] if high else [3, 4, 5, 6]
        y_data = [20, 50, 70, 90]

    elif key == "030":
        x_data = [2, 3, 7, 9] if high else [0, 1, 2, 3]
        y_data = [14, 12, 10, 8]
    else:
        raise ValueError("Unknown key", key)

    return LineData(
        key=key,
        line_params=LineParams(name=name, colour=colour),
        x=np.array(x_data) if x else np.array([]),
        y=np.array(y_data),
        default_indices=default_indices,
    )


@pytest.mark.parametrize(
    "name,plot_id,current_data,new_points_msg,expected",
    [
        (
            "equal_length_no_default",
            "plot_0",
            MultiLineDataMessage(
                ml_data=[generate_test_data("100"), generate_test_data("200")]
            ),
            AppendLineDataMessage(
                al_data=[generate_test_data("010"), generate_test_data("020")]
            ),
            (
                MultiLineDataMessage(
                    ml_data=[
                        generate_test_data("100", combined=True),
                        generate_test_data("200", combined=True),
                    ]
                ),
                AppendLineDataMessage(
                    al_data=[
                        generate_test_data("010", colour="#009e73"),
                        generate_test_data("020", colour="#e69d00"),
                    ]
                ),
            ),
        ),
        (
            "more_current_data_no_default",
            "plot_0",
            MultiLineDataMessage(
                ml_data=[
                    generate_test_data("100"),
                    generate_test_data("200", x=True, high=True),
                    generate_test_data("300", high=True),
                ]
            ),
            AppendLineDataMessage(
                al_data=[
                    generate_test_data("010"),
                    generate_test_data("020", high=True),
                ]
            ),
            (
                MultiLineDataMessage(
                    ml_data=[
                        generate_test_data("100", combined=True),
                        generate_test_data("200", combined=True, high=True),
                        generate_test_data("300", high=True),
                    ]
                ),
                AppendLineDataMessage(
                    al_data=[
                        generate_test_data("010", colour="#009e73"),
                        generate_test_data("020", high=True, colour="#e69d00"),
                    ]
                ),
            ),
        ),
        (
            "more_append_data_no_default",
            "plot_0",
            MultiLineDataMessage(
                ml_data=[
                    generate_test_data("100"),
                    generate_test_data("200", high=True),
                ]
            ),
            AppendLineDataMessage(
                al_data=[
                    generate_test_data("010"),
                    generate_test_data("020", high=True),
                    generate_test_data("030", high=True),
                ]
            ),
            (
                MultiLineDataMessage(
                    ml_data=[
                        generate_test_data("100", combined=True),
                        generate_test_data("200", combined=True, high=True),
                        generate_test_data("030", high=True, colour="#56b3e9"),
                    ]
                ),
                AppendLineDataMessage(
                    al_data=[
                        generate_test_data("010", colour="#009e73"),
                        generate_test_data("020", high=True, colour="#e69d00"),
                        generate_test_data("030", high=True, colour="#56b3e9"),
                    ]
                ),
            ),
        ),
        (
            "equal_length_default_indices",
            "plot_1",
            MultiLineDataMessage(
                ml_data=[
                    generate_test_data("100", default_indices=True, colour="#009e73"),
                    generate_test_data("200", default_indices=True, colour="#e69d00"),
                ]
            ),
            AppendLineDataMessage(
                al_data=[
                    generate_test_data("010", x=False),
                    generate_test_data("020", high=False),
                ]
            ),
            (
                MultiLineDataMessage(
                    ml_data=[
                        generate_test_data(
                            "100", default_indices=True, combined=True, colour="#009e73"
                        ),
                        generate_test_data(
                            "200", default_indices=True, combined=True, colour="#e69d00"
                        ),
                    ]
                ),
                AppendLineDataMessage(
                    al_data=[
                        generate_test_data(
                            "010", default_indices=True, colour="#009e73"
                        ),
                        generate_test_data(
                            "020", default_indices=True, colour="#e69d00"
                        ),
                    ]
                ),
            ),
        ),
        (
            "more_current_data_default_indices",
            "plot_0",
            MultiLineDataMessage(
                ml_data=[
                    generate_test_data("100", default_indices=True),
                    generate_test_data("200", default_indices=True),
                    generate_test_data("300", default_indices=True),
                ]
            ),
            AppendLineDataMessage(
                al_data=[
                    generate_test_data("010", x=False),
                    generate_test_data("020", x=False),
                ]
            ),
            (
                MultiLineDataMessage(
                    ml_data=[
                        generate_test_data("100", default_indices=True, combined=True),
                        generate_test_data("200", default_indices=True, combined=True),
                        generate_test_data("300", default_indices=True),
                    ]
                ),
                AppendLineDataMessage(
                    al_data=[
                        generate_test_data(
                            "010", default_indices=True, colour="#009e73"
                        ),
                        generate_test_data(
                            "020", default_indices=True, colour="#e69d00"
                        ),
                    ]
                ),
            ),
        ),
        (
            "more_current_data_default_indices_some_indices_given",
            "plot_0",
            MultiLineDataMessage(
                ml_data=[
                    generate_test_data("100", default_indices=True),
                    generate_test_data("200", default_indices=True),
                    generate_test_data("300", default_indices=True),
                ]
            ),
            AppendLineDataMessage(
                al_data=[
                    generate_test_data("010", x=False),
                    generate_test_data("020", high=True),
                ]
            ),
            (
                MultiLineDataMessage(
                    ml_data=[
                        generate_test_data("100", default_indices=True, combined=True),
                        generate_test_data("200", default_indices=True, combined=True),
                        generate_test_data("300", default_indices=True),
                    ]
                ),
                AppendLineDataMessage(
                    al_data=[
                        generate_test_data(
                            "010", default_indices=True, colour="#009e73"
                        ),
                        generate_test_data(
                            "020", default_indices=True, colour="#e69d00"
                        ),
                    ]
                ),
            ),
        ),
        (
            "more_append_data_default_indices_indices_given",
            "plot_0",
            MultiLineDataMessage(
                ml_data=[
                    generate_test_data("100", default_indices=True),
                    generate_test_data("200", default_indices=True),
                ]
            ),
            AppendLineDataMessage(
                al_data=[
                    generate_test_data("010"),
                    generate_test_data("020", high=True),
                    generate_test_data("030", high=True),
                ]
            ),
            (
                MultiLineDataMessage(
                    ml_data=[
                        generate_test_data("100", default_indices=True, combined=True),
                        generate_test_data("200", default_indices=True, combined=True),
                        generate_test_data(
                            "030", default_indices=True, colour="#56b3e9"
                        ),
                    ]
                ),
                AppendLineDataMessage(
                    al_data=[
                        generate_test_data(
                            "010", default_indices=True, colour="#009e73"
                        ),
                        generate_test_data(
                            "020", default_indices=True, colour="#e69d00"
                        ),
                        generate_test_data(
                            "030", default_indices=True, colour="#56b3e9"
                        ),
                    ]
                ),
            ),
        ),
        (
            "more_append_data_default_indices",
            "plot_0",
            MultiLineDataMessage(
                ml_data=[
                    generate_test_data("100", default_indices=True),
                    generate_test_data("200", default_indices=True),
                ]
            ),
            AppendLineDataMessage(
                al_data=[
                    generate_test_data("010", x=False),
                    generate_test_data("020", x=False),
                    generate_test_data("030", x=False),
                ]
            ),
            (
                MultiLineDataMessage(
                    ml_data=[
                        generate_test_data("100", default_indices=True, combined=True),
                        generate_test_data("200", default_indices=True, combined=True),
                        generate_test_data(
                            "030", default_indices=True, colour="#56b3e9"
                        ),
                    ]
                ),
                AppendLineDataMessage(
                    al_data=[
                        generate_test_data(
                            "010", default_indices=True, colour="#009e73"
                        ),
                        generate_test_data(
                            "020", default_indices=True, colour="#e69d00"
                        ),
                        generate_test_data(
                            "030", default_indices=True, colour="#56b3e9"
                        ),
                    ]
                ),
            ),
        ),
    ],
)
def test_combine_line_messages(
    name: str,
    plot_id: str,
    current_data: DataMessage,
    new_points_msg: AppendLineDataMessage,
    expected: tuple[MultiLineDataMessage, AppendLineDataMessage],
):
    ps = PlotServer()
    ps.plot_states[plot_id].current_data = current_data
    ml_msg, al_msg = ps.combine_line_messages(plot_id, new_points_msg)

    assert_line_data_messages_are_equal(ml_msg, expected[0])
    assert_line_data_messages_are_equal(al_msg, expected[1])


@pytest.mark.expect_caplog_errors
@pytest.mark.asyncio
async def test_add_and_remove_clients(caplog):
    websocket_0 = Mock()
    ps = PlotServer()

    data_0 = {"a": 10, "b": 20}
    selection_0 = {"selection0": 30}
    msg_00 = ws_pack(data_0)
    msg_01 = ws_pack(selection_0)

    plot_state_0 = ps.plot_states["plot_0"]
    plot_state_0.new_data_message = msg_00
    plot_state_0.new_selections_message = msg_01
    plot_state_0.current_data = data_0  # pyright: ignore[reportGeneralTypeIssues]
    plot_state_0.current_selections = selection_0  # pyright: ignore[reportGeneralTypeIssues]

    def update_plot_state(pc, bytes):
        time.sleep(2)
        if not plot_state_0.lock.locked():
            plot_state_0.clear()

    assert ps.client_total == 0
    assert ps._clients == {}

    with before_after.after(
        "davidia.server.plotserver.PlotClient.add_message", update_plot_state
    ):
        pc_0 = await ps.add_client("plot_0", websocket_0, "7b2ee613")

    assert pc_0.name == "plot_0:0"
    assert pc_0.websocket == websocket_0
    assert pc_0.queue.qsize() == 3  # do not use queue.empty()
    assert ps.client_total == 1
    assert ps._clients["plot_0"] == [pc_0]

    ps.plot_states["plot_0"] = PlotState(msg_00, None, data_0, None)
    websocket_1 = Mock()
    pc_1 = await ps.add_client("plot_0", websocket_1, "743f8791")

    assert pc_1.name == "plot_0:1"
    assert pc_1.websocket == websocket_1
    assert pc_1.queue.qsize() == 2
    assert ps.client_total == 2
    assert ps._clients["plot_0"] == [pc_0, pc_1]

    await ps.remove_client("plot_0", pc_0)

    assert ps._clients == defaultdict(list, {"plot_0": [pc_1]})

    await ps.remove_client("plot_0", pc_0)
    assert ps._clients == defaultdict(list, {"plot_0": [pc_1]})
    assert "Client plot_0:0 does not exist for plot_0" in caplog.text


def test_get_plot_ids():
    ps = PlotServer()

    assert ps.get_plot_ids() == []

    ps._clients["plot_0"].append("0")  # pyright: ignore[reportGeneralTypeIssues]
    ps._clients["plot_1"].append("1")  # pyright: ignore[reportGeneralTypeIssues]

    assert ps.get_plot_ids() == ["plot_0", "plot_1"]


@pytest.mark.asyncio
async def test_regions():
    regions = [
        LinearSelection(start=(1, -2.5), length=3.2, degrees=30, colour="green"),
        RectangularSelection(
            start=(-4.3, 5.6), lengths=(6, 7.8), degrees=-60, colour="red", alpha=0.2
        ),
    ]
    ps = PlotServer()
    pid = "plot_7"
    await ps.prepare_data(
        PlotMessage(
            plot_id=pid,
            type=MsgType.new_selection_data,
            params=SelectionsMessage(set_selections=regions),
        )
    )
    new_regions = await ps.get_regions(pid)
    assert len(regions) == len(new_regions)
    for e, a in zip(regions, new_regions):
        assert e == a

    new_region = RectangularSelection(
        start=(4.3, -5.6), lengths=(6, 7.8), degrees=0, colour="orange", alpha=0.7
    )
    await ps.prepare_data(
        PlotMessage(
            plot_id=pid,
            type=MsgType.update_selection_data,
            params=UpdateSelectionsMessage(update_selections=[new_region]),
        )
    )
    new_regions = await ps.get_regions(pid)
    expected_regions = list(regions)
    expected_regions.append(new_region)
    assert len(expected_regions) == len(new_regions)
    for e, a in zip(expected_regions, new_regions):
        assert e == a

    update = 2
    updated_region = expected_regions[update]
    updated_region.angle = 123.4
    await ps.prepare_data(
        PlotMessage(
            plot_id=pid,
            type=MsgType.update_selection_data,
            params=UpdateSelectionsMessage(update_selections=[new_region]),
        )
    )
    new_regions = await ps.get_regions(pid)
    expected_regions[update].angle = 123.4
    assert len(expected_regions) == len(new_regions)
    for e, a in zip(expected_regions, new_regions):
        assert e == a

    remove = 1
    await ps.prepare_data(
        PlotMessage(
            plot_id=pid,
            type=MsgType.clear_selection_data,
            params=ClearSelectionsMessage(selection_ids=[new_regions[remove].id]),
        )
    )
    new_regions = await ps.get_regions(pid)
    expected_regions.pop(remove)
    assert len(expected_regions) == len(new_regions)
    for e, a in zip(expected_regions, new_regions):
        assert e == a

    await ps.prepare_data(
        PlotMessage(
            plot_id=pid,
            type=MsgType.clear_selection_data,
            params=ClearSelectionsMessage(selection_ids=[]),
        )
    )
    new_regions = await ps.get_regions(pid)
    assert len(new_regions) == 0


def test_clients_available():
    ps = PlotServer()

    assert not ps.clients_available()

    ps._clients["plot_0"].append("0")  # pyright: ignore[reportGeneralTypeIssues]
    ps._clients["plot_1"].append("1")  # pyright: ignore[reportGeneralTypeIssues]

    assert ps.clients_available()


@pytest.mark.asyncio
async def test_clear_queues():
    ps = PlotServer()
    websocket_0 = AsyncMock()
    websocket_1 = AsyncMock()

    pc_0 = PlotClient(websocket_0, "fc1e5b22")
    pc_0.name = "plot_client_0"
    pc_1 = PlotClient(websocket_1, "fc1e5b22")
    pc_1.name = "plot_client_1"

    ps.client_total = 2
    ps._clients["plot_0"].append(pc_0)
    ps._clients["plot_1"].append(pc_1)

    data_0 = {"a": 10, "b": 20}
    selection_0 = {"selection0": 30}
    data_1 = {"e": 50, "f": 60}

    msg_00 = ws_pack(data_0)
    msg_01 = ws_pack(selection_0)
    msg_10 = ws_pack(data_1)

    ps.plot_states["plot_0"] = PlotState(msg_00, msg_01, data_0, selection_0)
    ps.plot_states["plot_1"] = PlotState(msg_10, None, data_1, None)

    plot_state_0 = ps.plot_states["plot_0"]
    plot_state_1 = ps.plot_states["plot_1"]

    await pc_0.queue.put(msg_00)
    await pc_0.queue.put(msg_01)
    await pc_1.queue.put(msg_10)

    assert plot_state_0.new_data_message == msg_00
    assert plot_state_0.new_selections_message == msg_01
    assert plot_state_0.current_data == data_0
    assert plot_state_0.current_selections == selection_0

    assert plot_state_1.new_data_message == msg_10
    assert plot_state_1.new_selections_message is None
    assert plot_state_1.current_data == data_1
    assert plot_state_1.current_selections is None

    assert pc_0.queue.qsize() != 0
    assert pc_1.queue.qsize() != 0

    await ps.clear_queues("plot_0")

    assert plot_state_0.new_data_message is None
    assert plot_state_0.new_selections_message is None
    assert plot_state_0.current_data is None
    assert plot_state_0.current_selections is None

    assert plot_state_1.new_data_message == msg_10
    assert plot_state_1.new_selections_message is None
    assert plot_state_1.current_data == data_1
    assert plot_state_1.current_selections is None

    assert pc_0.queue.qsize() == 0
    assert pc_1.queue.qsize() != 0


@pytest.mark.asyncio
async def test_clear_plot_states():
    data_1 = {"e": 50, "f": 60}
    msg_10 = ws_pack(data_1)
    ps = PlotServer()
    ps.plot_states["plot_1"] = PlotState(msg_10, None, data_1, None)
    plot_state_1 = ps.plot_states["plot_1"]

    def add_current_data(plotserver: PlotServer, plot_id: str):
        if not plot_state_1.lock.locked():
            plot_state_1.current_data = {
                "a": 10,
                "b": 20,
            }  # pyright: ignore[reportGeneralTypeIssues]

    with before_after.after(
        "davidia.server.plotserver.PlotServer.clear_plot_states", add_current_data
    ):
        await ps.clear_queues("plot_1")

    assert plot_state_1.current_data is None


@pytest.mark.asyncio
async def test_prepare_data():
    ps = PlotServer()
    websocket_0 = AsyncMock()
    pc_0 = PlotClient(websocket_0, "fc1e5b22")
    pc_0.name = "plot_client_0"
    ps.client_total = 1
    ps._clients["plot_0"].append(pc_0)
    data_0 = MultiLineDataMessage(
        ml_data=[
            LineData(
                key="",
                line_params=LineParams(),
                x=np.array([0, 1, 2]),
                y=np.array([0, 30, 20]),
            )
        ]
    )
    selection_0 = {"selection0": 30}
    msg_00 = ws_pack(data_0)
    msg_01 = ws_pack(selection_0)
    append_line = PlotMessage(
        plot_id="plot_0",
        type=MsgType.append_line_data,
        params=[
            {
                "line_params": LineParams(colour="purple"),
                "x": np.array([3, 4, 5]),
                "y": np.array([10, 20, 30]),
            }
        ],
    )

    plot_state_0 = ps.plot_states["plot_0"]
    plot_state_0 = PlotState(msg_00, msg_01, data_0, selection_0)

    assert plot_state_0.new_data_message == msg_00
    assert plot_state_0.new_selections_message == msg_01
    assert plot_state_0.current_data == data_0
    assert plot_state_0.current_selections == selection_0

    def change_plot_states(plotserver, plot_id, new_points_msg):
        if not plot_state_0.lock.locked():
            ta_msg = TableDataMessage(
                ta_data=TableData(
                    key="", cell_values=np.array([[1, 2], [3, 4]]), cell_width=120
                )
            )
            plot_state_0.current_data = ta_msg

    with before_after.after(
        "davidia.server.plotserver.PlotServer.combine_line_messages", change_plot_states
    ):
        await ps.prepare_data(append_line)
        assert isinstance(plot_state_0.current_data, MultiLineDataMessage)


@pytest.mark.parametrize(
    "name,msg,expected",
    [
        (
            "no_default_indices",
            MultiLineDataMessage(
                plot_config=PlotConfig(),
                ml_data=[generate_test_data("100"), generate_test_data("200")],
            ),
            MultiLineDataMessage(
                plot_config=PlotConfig(),
                ml_data=[generate_test_data("100"), generate_test_data("200")],
            ),
        ),
        (
            "default_indices",
            MultiLineDataMessage(
                plot_config=PlotConfig(),
                ml_data=[
                    generate_test_data("100", x=False),
                    generate_test_data("030", x=False),
                ],
            ),
            MultiLineDataMessage(
                plot_config=PlotConfig(),
                ml_data=[
                    generate_test_data("100", default_indices=True),
                    generate_test_data("030", default_indices=True),
                ],
            ),
        ),
    ],
)
def test_add_indices(name, msg: MultiLineDataMessage, expected: MultiLineDataMessage):
    message = add_indices(msg)

    assert_line_data_messages_are_equal(message, expected)


@pytest.mark.parametrize(
    "name,msg,expected",
    [
        (
            "no_default_indices",
            AppendLineDataMessage(
                plot_config=PlotConfig(),
                al_data=[generate_test_data("100"), generate_test_data("200")],
            ),
            MultiLineDataMessage(
                plot_config=PlotConfig(),
                ml_data=[
                    generate_test_data("100", colour="#009e73"),
                    generate_test_data("200", colour="#e69d00"),
                ],
            ),
        ),
        (
            "all_default_indices",
            AppendLineDataMessage(
                plot_config=PlotConfig(),
                al_data=[
                    generate_test_data("100", x=False),
                    generate_test_data("030", x=False),
                ],
            ),
            MultiLineDataMessage(
                plot_config=PlotConfig(),
                ml_data=[
                    generate_test_data("100", default_indices=True, colour="#009e73"),
                    generate_test_data("030", default_indices=True, colour="#e69d00"),
                ],
            ),
        ),
        (
            "some_default_indices",
            AppendLineDataMessage(
                plot_config=PlotConfig(),
                al_data=[generate_test_data("100"), generate_test_data("030", x=False)],
            ),
            MultiLineDataMessage(
                plot_config=PlotConfig(),
                ml_data=[
                    generate_test_data("100", default_indices=True, colour="#009e73"),
                    generate_test_data("030", default_indices=True, colour="#e69d00"),
                ],
            ),
        ),
    ],
)
def test_convert_append_to_multi_line_data_message(
    name, msg: AppendLineDataMessage, expected: MultiLineDataMessage
):
    message = convert_append_to_multi_line_data_message(msg)

    assert_line_data_messages_are_equal(message, expected)


@pytest.mark.parametrize(
    "name,input,expected",
    [
        (
            "all_lines_named",
            [
                generate_test_data(key="100", name="first"),
                generate_test_data(key="200", name="second"),
                generate_test_data(key="300", name="third"),
            ],
            [
                generate_test_data(key="100", name="first"),
                generate_test_data(key="200", name="second"),
                generate_test_data(key="300", name="third"),
            ],
        ),
        (
            "no_lines_named",
            [
                generate_test_data(key="100"),
                generate_test_data(key="200"),
                generate_test_data(key="300"),
            ],
            [
                generate_test_data(key="100", name="Line 0"),
                generate_test_data(key="200", name="Line 1"),
                generate_test_data(key="300", name="Line 2"),
            ],
        ),
        (
            "empty list",
            [],
            [],
        ),
        (
            "single_line_data",
            [generate_test_data(key="100", name="first")],
            [generate_test_data(key="100", name="first")],
        ),
        (
            "original_names_repeated",
            [
                generate_test_data(key="100", name="first"),
                generate_test_data(key="200", name="first"),
                generate_test_data(key="300", name="third"),
            ],
            [
                generate_test_data(key="100", name="first"),
                generate_test_data(key="200", name="first"),
                generate_test_data(key="300", name="third"),
            ],
        ),
        (
            "names_will_repeat",
            [
                generate_test_data(key="100", name="first"),
                generate_test_data(key="200", name="Line 1"),
                generate_test_data(key="300"),
            ],
            [
                generate_test_data(key="100", name="first"),
                generate_test_data(key="200", name="Line 1"),
                generate_test_data(key="300", name="Line 0"),
            ],
        ),
        (
            "multiple_names_will_repeat",
            [
                generate_test_data(key="100", name="Line 1"),
                generate_test_data(key="200"),
                generate_test_data(key="300"),
                generate_test_data(key="010", name="Line 0"),
                generate_test_data(key="020"),
            ],
            [
                generate_test_data(key="100", name="Line 1"),
                generate_test_data(key="200", name="Line 2"),
                generate_test_data(key="300", name="Line 3"),
                generate_test_data(key="010", name="Line 0"),
                generate_test_data(key="020", name="Line 4"),
            ],
        ),
    ],
)
def test_check_line_names(name, input: list[LineData], expected: list[LineData]):
    renamed_lines = check_line_names(input)

    assert all([line_data_are_equal(a, b) for a, b in zip(expected, renamed_lines)])



import logging

import requests
from davidia.server.benchmarks import BENCHMARK_HELP, BenchmarkParams, PlotType

logger = logging.getLogger("benchmark")


def start_benchmark(
    params: BenchmarkParams,
    plot_id: str = "plot_0",
    host: str = "127.0.0.1",
    port: int = 8000,
) -> requests.Response:
    """Start benchmark

    Parameters
    ----------
    params: BenchmarkParams
    plot_id: str

    Returns
    -------
    response: Response
        Response from push_data POST request
    """

    url = f"http://{host}:{port}/benchmark/{plot_id}"
    logger.debug("Sending POST to %s: %s", url, params)
    response = requests.post(url, data=params.model_dump_json())

    logger.debug(
        f"{params} plotted in {response.elapsed}s with response"
        f" status code is {response.status_code}.\n"
    )

    return response


def create_parser():
    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

    parser = ArgumentParser(
        description="Benchmark plotting client",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    plot_types = list(p.name for p in PlotType)
    parser.add_argument("-i", "--id", help="Plot ID", default="plot_0")
    parser.add_argument(
        "-t", "--type", help="Plot type", choices=plot_types, default=plot_types[0]
    )
    parser.add_argument(
        "-r", "--repetitions", help="Number of repeats", type=int, default=1
    )
    parser.add_argument(
        "-p",
        "--pause",
        help="Period to pause between changes (in seconds)",
        type=float,
        default=0.2,
    )
    parser.add_argument(
        "-H", "--host", help="Set the host address for server", default="127.0.0.1"
    )
    parser.add_argument(
        "-P", "--port", help="Set the port number for server", type=int, default=8000
    )
    plot_help = [
        f"{n}: {h}"
        for n, h in ((p.name, BENCHMARK_HELP.get(p, None)) for p in PlotType)
        if h is not None
    ]
    parser.add_argument("params", nargs="*", help="; ".join(plot_help), type=float)
    return parser


def main():
    args = create_parser().parse_args()
    response = start_benchmark(
        BenchmarkParams(
            plot_type=args.type,
            params=args.params,
            iterations=args.repetitions,
            pause=args.pause,
        ),
        plot_id=args.id,
        host=args.host,
        port=args.port,
    )
    print(f"{response.status_code}: {response.content.decode()}")


if __name__ == "__main__":
    logger.setLevel(logging.DEBUG)
    main()



"""colourMap str options are listed in INTERPOLATORS in h5web https://github.com/silx-kit/h5web/blob/main/packages/lib/src/vis/heatmap/interpolators.ts
ScaleType enum options are linear, log, symlog, sqrt, gamma
"""

from __future__ import annotations

import logging
import warnings
from time import time_ns
from typing import Any

import numpy as np
import requests
from davidia.models.messages import (
    ClearSelectionsMessage,
    ColourMap,
    GlyphType,
    HeatmapData,
    ImageData,
    LineData,
    LineParams,
    MsgType,
    PlotMessage,
    ScaleType,
    ScatterData,
    SelectionsMessage,
    SurfaceData,
    TableData,
    UpdateSelectionsMessage,
)
from davidia.models.parameters import PlotConfig, TableDisplayParams, TableDisplayType
from davidia.models.selections import (
    AnySelection,
    AxialSelection,
    CircularSectorialSelection,
    CircularSelection,
    EllipticalSelection,
    LinearSelection,
    PolygonalSelection,
    RectangularSelection,
)
from davidia.server.fastapi_utils import j_dumps, j_loads, ws_pack
from numpy.typing import ArrayLike

OptionalArrayLike = ArrayLike | None
OptionalLists = OptionalArrayLike | list[OptionalArrayLike] | None

Selections = AnySelection | list[AnySelection]


class PlotConnection:
    def __init__(self, plot_id, host="localhost", port=8000, use_msgpack=True):
        self.plot_id = plot_id
        self.host = host
        self.port = port
        self.url_prefix = f"http://{host}:{port}/"
        self.use_msgpack = use_msgpack

    def _prepare_request(self, data):
        if data is None:
            return None, None
        if self.use_msgpack:
            data = ws_pack(data)
            return data, {
                "Content-Type": "application/x-msgpack",
            }
        data = j_dumps(data)
        return data, None

    def _post(
        self,
        params,
        msg_type=MsgType.new_multiline_data,
        plot_config=None,
        endpoint="push_data",
    ):
        url = self.url_prefix + endpoint
        data = PlotMessage(
            plot_id=self.plot_id, type=msg_type, params=params, plot_config=plot_config
        )
        logging.debug("posting PM: %s", data)
        start = time_ns()
        data, headers = self._prepare_request(data)
        resp = requests.post(url, data=data, headers=headers)
        elapsed = (time_ns() - start) // 1000000
        logging.info("plot_server.post %d, %dms", resp.status_code, elapsed)
        return resp

    def _put(self, data, endpoint):
        url = self.url_prefix + endpoint
        start = time_ns()
        data, headers = self._prepare_request(data)
        resp = requests.put(url, data=data, headers=headers)
        elapsed = (time_ns() - start) // 1000000
        logging.info("plot_server.put %d, %dms", resp.status_code, elapsed)
        return resp

    def _get(self, endpoint):
        url = self.url_prefix + endpoint
        start = time_ns()
        resp = requests.get(url)
        elapsed = (time_ns() - start) // 1000000
        logging.info("plot_server.get %d, %dms", resp.status_code, elapsed)
        return resp

    def get_plots_ids(self) -> list[str]:
        ids = j_loads(self._get("get_plot_ids").content)
        if isinstance(ids, list):
            return ids
        logging.warning("Fetched values not a list (%s): %s", type(ids), ids)
        return []

    @staticmethod
    def _as_list(item_list, n):
        if isinstance(item_list, list):
            n_items = len(item_list)
            if n_items < n:
                logging.warning("Supplied list is too short: %d cf %d", n_items, n)
                return item_list * (n // n_items) + item_list[: (n % n_items)]
            return item_list
        return [item_list] * n

    @staticmethod
    def _populate_plot_config(
        plot_config: dict[str, Any] | None,
        x: OptionalArrayLike = None,
        y: OptionalArrayLike = None,
    ) -> PlotConfig:
        plot_config = dict(plot_config) if plot_config is not None else {}

        if hasattr(plot_config, "x_values"):
            plot_config["x_values"] = np.asanyarray(plot_config["x_values"])
        if x is not None:
            plot_config["x_values"] = np.asanyarray(x)
        if hasattr(plot_config, "y_values"):
            plot_config["y_values"] = np.asanyarray(plot_config["y_values"])
        if y is not None:
            plot_config["y_values"] = np.asanyarray(y)
        return PlotConfig.model_validate(plot_config)

    def line(
        self,
        x: OptionalLists,
        y: OptionalLists = None,
        plot_config: dict[str, Any] | None = None,
        append: bool = False,
        **attribs,
    ):
        """Plot line

        Parameters
        ----------
        x: x (or y if y not given) array
        y: y array (if x given)
        plot_config: plot config
        append: add line to existing multiline plot
        attribs: dict of attributes for plot

        Returns
        -------
        response: Response
            Response from push_data POST request
        """
        if y is None:
            xf = None
            yf = x
        else:
            xf = x
            yf = y
        if yf is None:
            return

        if append:
            msg_type = MsgType.append_line_data
        else:
            msg_type = MsgType.new_multiline_data

        def _asanyarray(x):
            return x if x is None else np.asanyarray(x)

        global_attribs = dict(attribs)
        line_on = global_attribs.pop("line_on", True)
        glyph_type = global_attribs.pop("glyph_type", None)
        glyph_type = (
            glyph_type
            if glyph_type is None or isinstance(glyph_type, GlyphType)
            else GlyphType[glyph_type]
        )
        colour = global_attribs.pop("colour", None)
        point_size = global_attribs.pop("point_size", None)

        if isinstance(yf, list) and isinstance(yf[0], (list, np.ndarray)):
            n_plots = len(yf)
            xl: list[OptionalArrayLike] = []
            if xf is None:
                xl = [[]] * n_plots
            elif isinstance(xf, np.ndarray):
                xl = [xf] * n_plots
            elif isinstance(xf, list):
                if isinstance(xf[0], (list, np.ndarray)):
                    n_x = len(xf)
                    if n_x == 1:
                        xl = [_asanyarray(xf[0])] * n_plots
                    elif n_x < n_plots:
                        raise ValueError("Number of x arrays must be match y arrays")
                    else:
                        xl = [_asanyarray(xi) for xi in xf]
                else:
                    xl = [_asanyarray(xf)] * n_plots

            lines_on = PlotConnection._as_list(line_on, n_plots)
            glyph_types = PlotConnection._as_list(glyph_type, n_plots)
            glyph_types = [
                g if g is None or isinstance(g, GlyphType) else GlyphType[g]
                for g in glyph_types
            ]
            colours = PlotConnection._as_list(colour, n_plots)
            point_sizes = PlotConnection._as_list(point_size, n_plots)
            plot_config = PlotConnection._populate_plot_config(plot_config)
            lds = []
            for xi, yi, ci, li, ps, gt in zip(
                xl, yf, colours, lines_on, point_sizes, glyph_types
            ):
                if yi is None:
                    raise ValueError("The y data must not contain None")
                line_params = LineParams(
                    colour=ci,
                    line_on=li,
                    point_size=ps,
                    glyph_type=gt,
                )

                lds.append(
                    LineData(
                        line_params=line_params,
                        x=_asanyarray(xi),
                        y=_asanyarray(yi),
                        **global_attribs,
                    )
                )
        else:
            lds = [
                LineData(
                    line_params=LineParams(
                        colour=colour,
                        line_on=line_on,
                        point_size=point_size,
                        glyph_type=glyph_type,
                    ),
                    x=_asanyarray(xf),
                    y=_asanyarray(yf),
                    **global_attribs,
                )
            ]
        return self._post(lds, msg_type=msg_type, plot_config=plot_config)

    def image(
        self,
        image: OptionalLists,
        x: OptionalArrayLike = None,
        y: OptionalArrayLike = None,
        plot_config: dict[str, Any] | None = None,
        **attribs,
    ):
        """Plot image

        Parameters
        ----------
        image: array
        x: x array
        y: y array
        plot_config: plot config
        Returns
        -------
        response: Response
            Response from push_data POST request
        """
        values = np.asanyarray(image)
        if values.ndim == 2:
            if "domain" not in attribs:
                attribs["domain"] = values.min(), values.max()
            im = HeatmapData(values=values, **attribs)
        elif values.ndim == 3 and values.shape[2] == 3:
            im = ImageData(values=values, **attribs)
        else:
            raise ValueError("Data cannot be interpreted as heatmap or image data")
        plot_config = PlotConnection._populate_plot_config(plot_config, x, y)
        return self._post(im, msg_type=MsgType.new_image_data, plot_config=plot_config)

    def scatter(
        self,
        x: ArrayLike,
        y: ArrayLike,
        point_values: OptionalLists,
        domain: tuple[float, float],
        plot_config: dict[str, Any] | None = None,
        **attribs,
    ):
        """Plot scatter data
        Parameters
        ----------
        x: x coordinates
        y: y coordinates
        point_values: array
        domain: tuple
        plot_config: plot config
        Returns
        -------
        response: Response
            Response from push_data POST request
        """
        sc = ScatterData(
            x=np.asanyarray(x),
            y=np.asanyarray(y),
            point_values=np.asanyarray(point_values),
            domain=domain,
            **attribs,
        )
        plot_config = PlotConnection._populate_plot_config(plot_config)
        return self._post(
            sc, msg_type=MsgType.new_scatter_data, plot_config=plot_config
        )

    def surface(
        self,
        values: OptionalLists,
        domain: tuple[float, float],
        x: OptionalArrayLike = None,
        y: OptionalArrayLike = None,
        plot_config: dict[str, Any] | None = None,
        **attribs,
    ):
        """Plot surface

        Parameters
        ----------
        values: array
        x: x array
        y: y array
        domain: tuple
        plot_config: plot config
        Returns
        -------
        response: Response
            Response from push_data POST request
        """
        su = SurfaceData(height_values=np.asanyarray(values), domain=domain, **attribs)
        plot_config = PlotConnection._populate_plot_config(plot_config, x, y)
        return self._post(
            su, msg_type=MsgType.new_surface_data, plot_config=plot_config
        )

    def table(
        self,
        data: OptionalLists,
        cell_width: int,
        display_style: TableDisplayType | None = None,
        number_digits: int | None = None,
        title: str | None = None,
        **attribs,
    ):
        """Show table of data

        Parameters
        ----------
        data: array
        cell_width: int
        display_style: notation type for displaying data
        number_digits: number significant or fractional digits to display
        title: title of plot
        Returns
        -------
        response: Response
            Response from push_data POST request
        """
        ta = TableData(
            cell_values=np.asanyarray(data),
            cell_width=cell_width,
            display_params=TableDisplayParams(
                display_type=display_style, number_digits=number_digits
            ),
            **attribs,
        )
        return self._post(ta, msg_type=MsgType.new_table_data)

    def clear(self) -> requests.Response:
        """Sends request to clear a plot

        Returns
        -------
        response: Response
            Response from clear_data GET request
        """
        return self._put(None, f"clear_data/{self.plot_id}")

    def region(
        self,
        selections: Selections | None,
        update: bool = False,
        delete: bool | str | list[str] = False,
    ):
        """Get, set or delete regions of selection

        Parameters
        ----------
        selections : Selections | None
            a selection or list of selections to set or None to get them
        update : bool
            update or add to existing selections
        delete : bool | str | list[str]
            True - remove given selections or all selections if None given
            str | list - remove selection(s) with given ID(s)

        Returns
        -------
        response: Response
            Response from push_data POST request
        """
        if selections is None and delete is False:
            return j_loads(self._get(f"get_regions/{self.plot_id}").content)

        if selections and not isinstance(selections, (tuple, list)):
            selections = [selections]

        if delete:
            if isinstance(delete, str):
                remove = [delete]
            elif isinstance(delete, (tuple, list)):
                remove = list(delete)
            elif selections is None:
                remove = []
            else:
                remove = [s.id for s in selections]
            msg_type = MsgType.clear_selection_data
            sm = ClearSelectionsMessage(selection_ids=remove)
        elif update and selections is not None:
            msg_type = MsgType.update_selection_data
            sm = UpdateSelectionsMessage(update_selections=selections)
        elif selections is not None:
            msg_type = MsgType.new_selection_data
            sm = SelectionsMessage(set_selections=selections)
        else:
            raise ValueError("Should not be reached")
        return self._post(sm, msg_type=msg_type)


_ALL_PLOTS: dict[str, PlotConnection] = dict()
_DEF_PLOT_ID = None

_DEF_PS_HOST = "localhost"
_DEF_PS_PORT = 8000


def set_default_plot_server(host: str, port: int):
    """Set default host and port for plot server

    Parameters
    ----------
    host : str
        host as IP address or name
    port : int
        port number
    """
    global _DEF_PS_HOST, _DEF_PS_PORT
    _DEF_PS_HOST = host
    _DEF_PS_PORT = port


def get_plot_connection(plot_id="", host=None, port=None):
    """Get a connection to plot server that has plot with given ID

    Parameters
    ----------
    plot_id : str
        if empty then creates a connection
    host : str
        name or IP of plot server
    port : int
        port number of plot server
    Returns
    -------
        instance of PlotConnection

    Side effects
    ------------
    Populates a global dict/cache that maps plot ID to connection
    Sets default plot ID from given plot ID or first ID on server
    """
    if plot_id and plot_id in _ALL_PLOTS:
        return _ALL_PLOTS[plot_id]

    if host is None:
        host = _DEF_PS_HOST
    if port is None:
        port = _DEF_PS_PORT
    pc = PlotConnection(plot_id, host, port)
    ids = pc.get_plots_ids()
    if len(ids) == 0:
        raise ValueError("Plot connection has no plots")
    if plot_id and plot_id not in ids:
        warnings.warn(f"Plot connection does not contain {plot_id}", Warning)

    global _DEF_PLOT_ID
    if plot_id:
        if plot_id in _ALL_PLOTS and pc is not _ALL_PLOTS[plot_id]:
            logging.warning(
                "Plot ID %s already exists in another connection, replacing"
                " with new connection",
                plot_id,
            )
        _DEF_PLOT_ID = plot_id
    else:
        _DEF_PLOT_ID = ids[0]
        pc = PlotConnection(_DEF_PLOT_ID, host, port)
    _ALL_PLOTS[_DEF_PLOT_ID] = pc
    return pc


def set_default_plot_id(plot_id: str | None):
    if not plot_id:
        raise ValueError("Plot ID must not be None or empty")
    get_plot_connection(plot_id)


def _get_default_plot_id(plot_id: str | None = None):
    if plot_id:
        return plot_id
    if _DEF_PLOT_ID is None:
        get_plot_connection()
        if _DEF_PLOT_ID is None:
            raise ValueError("No default plot ID - please specify as parameter")
    return _DEF_PLOT_ID


def line(
    x: OptionalLists,
    y: OptionalLists = None,
    plot_config: dict[str, Any] | None = None,
    plot_id: str | None = None,
    append: bool = False,
    **attribs,
):
    """Plot line

    Parameters
    ----------
    x: x (or y if y not given) array
    y: y array (if x given)
    plot_config: plot config
    plot_id: ID of plot where line is added
    append: add line to existing multiline plot
    **attribs: keywords specific to line
    Keyword options for attribs are
    {
        "name": str
        "colour": str | None  # str is a CSS color value https://developer.mozilla.org/en-US/docs/Web/CSS/color_value
        "line_on": bool
        "point_size": int | None
        "glyph_type": str | GlyphType  # enum options are Circle, Cross, Square, Cap
    }

    Returns
    -------
    response: Response
        Response from push_data POST request
    """
    plot_id = _get_default_plot_id(plot_id)
    pc = get_plot_connection(plot_id)
    return pc.line(x, y, plot_config, append, **attribs)


def image(
    values: OptionalLists,
    x: OptionalArrayLike = None,
    y: OptionalArrayLike = None,
    plot_config: dict | None = None,
    plot_id: str | None = None,
    **attribs,
):
    """Plot image

    Parameters
    ----------
    values: array
    x: x array
    y: y array
    plot_config: plot config
    plot_id: ID of plot where image is added
    **attribs: keywords specific to image
    Keyword options for attribs are
    {
        "aspect": Aspect | float | int | None  # Aspect enum options are auto, equal
        "domain": tuple[float, float]
        "heatmap_scale": ScaleType
        "colour_map": str | ColourMap
    }

    Returns
    -------
    response: Response
        Response from push_data POST request
    """
    plot_id = _get_default_plot_id(plot_id)
    pc = get_plot_connection(plot_id)
    return pc.image(values, x, y, plot_config, **attribs)


def scatter(
    x: ArrayLike,
    y: ArrayLike,
    point_values: OptionalLists,
    domain: tuple[float, float],
    plot_config: dict | None = None,
    plot_id: str | None = None,
    **attribs,
):
    """Plot scatter data
    Parameters
    ----------
    x: x coordinates
    y: y coordinates
    point_values: array
    domain: tuple
    plot_config: plot config
    plot_id: ID of plot where scatter points are added
    **attribs: keywords specific to scatter
    Keyword options for attribs are
    {
        "colour_map": str | ColourMap
        "point_size": float
    }

    Returns
    -------
    response: Response
        Response from push_data POST request
    """
    plot_id = _get_default_plot_id(plot_id)
    pc = get_plot_connection(plot_id)
    return pc.scatter(x, y, point_values, domain, plot_config, **attribs)


def surface(
    values: OptionalLists,
    domain: tuple[float, float],
    x: OptionalArrayLike = None,
    y: OptionalArrayLike = None,
    plot_config: dict | None = None,
    plot_id: str | None = None,
    **attribs,
):
    """Plot surface

    Parameters
    ----------
    values: array
    domain: tuple
    x: x array
    y: y array
    plot_config: plot config
    plot_id: ID of plot where surface is added
    **attribs: keywords specific to surface
    Keyword options for attribs are
    {
        "surface_scale": str | ScaleType,
        "colour_map": str | ColourMap
    }

    Returns
    -------
    response: Response
        Response from push_data POST request
    """
    plot_id = _get_default_plot_id(plot_id)
    pc = get_plot_connection(plot_id)
    return pc.surface(values, domain, x, y, plot_config, **attribs)


def table(
    data: OptionalLists,
    cell_width: int = 120,
    display_style: TableDisplayType | None = None,
    number_digits: int | None = None,
    title: str | None = None,
    plot_id: str | None = None,
    **attribs,
):
    """Show table of data
    Parameters
    ----------
    data: array
    cell_width: int
    display_style: notation type for displaying data
    number_digits: number significant or fractional digits to display
    title: title of plot
    plot_id: ID of plot where as table is shown
    **attribs: keywords specific to table
    Returns
    -------
    response: Response
        Response from push_data POST request
    """
    plot_id = _get_default_plot_id(plot_id)
    pc = get_plot_connection(plot_id)
    return pc.table(data, cell_width, display_style, number_digits, title, **attribs)


def clear(plot_id: str | None = None):
    """Sends request to clear a plot

    Parameters
    ----------
    plot_id : str
        the plot from which to clear data

    Returns
    -------
    response: Response
        Response from clear_data GET request
    """
    plot_id = _get_default_plot_id(plot_id)
    pc = get_plot_connection(plot_id)
    return pc.clear()


def region(
    selections: Selections | None = None,
    update: bool = False,
    delete: bool | str | list[str] = False,
    plot_id: str | None = None,
):
    """Get, set or delete regions of selection on plot

    Parameters
    ----------
    selections : Selections | None
        a selection or list of selections to set or None to get them
    update : bool
        update or add to existing selections
    delete : bool | str | list[str]
        True - remove given selections or all selections if None given
        str | list - remove selection(s) with given ID(s)
    plot_id : str
        the plot from which to clear data

    Returns
    -------
    response: Response
        Response from push_data POST request
    """
    plot_id = _get_default_plot_id(plot_id)
    pc = get_plot_connection(plot_id)
    return pc.region(selections, update, delete)


__all__ = [  # pyright: ignore[reportUnsupportedDunderAll]
    PlotConnection,
    get_plot_connection,
    set_default_plot_id,
    AxialSelection,
    LinearSelection,
    RectangularSelection,
    PolygonalSelection,
    CircularSelection,
    EllipticalSelection,
    CircularSectorialSelection,
    ColourMap,
    GlyphType,
    ScaleType,
    TableDisplayType,
    line,
    image,
    scatter,
    surface,
    table,
    region,
    clear,
]



from __future__ import annotations

import logging
from asyncio import Lock, Queue, QueueEmpty, sleep
from collections import defaultdict
from time import time_ns

import numpy as np
from fastapi import WebSocket, WebSocketDisconnect
from pydantic import ValidationError

from . import benchmarks as _benchmark
from ..models.messages import (
    AppendLineDataMessage,
    BatonRequestMessage,
    BatonMessage,
    ClearPlotsMessage,
    ClearSelectionsMessage,
    ClientLineParametersMessage,
    ClientScatterParametersMessage,
    ColourMap,
    DataMessage,
    DvDNDArray,
    HeatmapData,
    ImageDataMessage,
    LineData,
    MsgType,
    MultiLineDataMessage,
    PlotMessage,
    SelectionMessage,
    SelectionsMessage,
    ScatterData,
    ScatterDataMessage,
    StatusType,
    SurfaceData,
    SurfaceDataMessage,
    UpdateSelectionsMessage,
)
from ..models.selections import SelectionBase
from .fastapi_utils import ws_pack, ws_unpack
from .processor import Processor

logger = logging.getLogger("main")

COLOURLIST = [
    "#009e73",  # teal
    "#e69d00",  # orange
    "#56b3e9",  # light blue
    "#f0e442",  # yellow
    "#0072b2",  # blue
    "#d55e00",  # dark orange
    "#cc79a7",  # pink
]


def add_colour_to_lines(line_data: list[LineData]):
    for i, line in enumerate(line_data):
        if not line.line_params.colour:
            line_data[i].line_params.colour = COLOURLIST[i % len(COLOURLIST)]


class PlotClient:
    """A class to represent a Web UI client that plots

    This manages a queue of messages to send to the client
    """

    def __init__(self, websocket: WebSocket, uuid: str):
        self.websocket = websocket
        self.uuid = uuid
        self.queue = Queue()
        self.name = ""

    async def add_message(self, message: bytes):
        """Add message for client"""
        logger.debug(
            "New message being added to client %s with name %s", self.uuid, self.name
        )
        await self.queue.put(message)

    def clear_queue(self):
        """Clear messages in client queue"""
        q = self.queue
        while not q.empty():
            q.get_nowait()
            q.task_done()

    async def send_next_message(self):
        """Send next message in queue to client"""
        try:
            msg = self.queue.get_nowait()
            self.queue.task_done()
            await self.websocket.send_bytes(msg)
        except QueueEmpty:
            logger.debug("Queue for websocket %s is empty", self.websocket)


class PlotState:
    """Class for representing the state of a plot"""

    def __init__(
        self,
        new_data_message=None,
        new_selections_message=None,
        current_data=None,
        current_selections=None,
        new_baton_message=None,
        current_baton=None,
    ):
        self.new_data_message: bytes | None = new_data_message
        self.new_selections_message: bytes | None = new_selections_message
        self.new_baton_message: bytes | None = new_baton_message
        self.current_data: DataMessage | None = current_data
        self.current_selections: list[SelectionBase] | None = current_selections
        self.current_baton: str | None = current_baton
        self.lock = Lock()

    def clear(self):
        """Clear all current and new data and selections"""
        self.new_data_message = None
        self.new_selections_message = None
        self.current_data = None
        self.current_selections = None


class PlotServer:
    """
    A class used to create a plot server that manages plot clients and messaging

    Attributes
    ----------
    processor : Processor
        The data processor.
    _clients : dict[str, list[PlotClient]]
        A dictionary containing all plot clients per plot ID.
    client_status : StatusType
        The status of the client.
    baton: str | None
        Current baton uuid
    plot_states : dict[str, PlotState] = defaultdict(PlotState)
        A dictionary containing plot states per plot_id
    client_total : int
        Number of clients added to server
    """

    def __init__(self):
        self.processor: Processor = Processor()
        self._clients: dict[str, list[PlotClient]] = defaultdict(list)
        self.client_status: StatusType = StatusType.busy
        self.uuids: list[str] = []
        self.baton: str | None = None
        self.plot_states: dict[str, PlotState] = defaultdict(PlotState)
        self.client_total = 0
        self.last_colour_maps: dict[str, ColourMap] = defaultdict(
            lambda: ColourMap.Greys
        )

    async def add_client(
        self, plot_id: str, websocket: WebSocket, uuid: str
    ) -> PlotClient:
        """Add a client given by a plot ID and websocket
        Parameters
        ----------
        plot_id : str
        websocket: WebSocket

        Returns the added client
        """
        client = PlotClient(websocket, uuid)
        client.name = f"{plot_id}:{self.client_total}"
        self.client_total += 1
        self._clients[plot_id].append(client)
        if uuid not in self.uuids:
            self.uuids.append(uuid)

        if plot_id in self.plot_states:
            plot_state = self.plot_states[plot_id]
            async with plot_state.lock:
                if plot_state.new_data_message:
                    await client.add_message(plot_state.new_data_message)
                if plot_state.new_selections_message:
                    await client.add_message(plot_state.new_selections_message)
        if not self.baton:
            self.baton = uuid
            logger.info("Baton updated to %s", self.baton)
        await self.update_baton()
        return client

    async def update_baton(self):
        """Updates plot state and sends messages for new baton"""
        processed_msg = BatonMessage(baton=self.baton, uuids=self.uuids)
        logger.debug("Baton updated: %s", processed_msg)
        for p, cl in self._clients.items():
            msg = await self.update_plot_states_with_message(processed_msg, p)
            if msg is None:
                continue
            for c in cl:
                await c.add_message(msg)

    def _any_client(self, plot_id: str, uuid: str) -> bool:
        for p, cl in self._clients.items():
            if p != plot_id:
                for c in cl:
                    if c.uuid == uuid:
                        return True
        return False

    async def remove_client(self, plot_id: str, client: PlotClient) -> bool:
        """Remove a client given by a plot ID and client
        Parameters
        ----------
        plot_id : str
        client : PlotClient

        Returns
        -------
        True if messages updated
        """
        try:
            self._clients[plot_id].remove(client)
        except ValueError:
            logger.warning("Client %s does not exist for %s", client.name, plot_id)

        uuid = client.uuid
        if self._any_client(plot_id, uuid):
            return False

        if uuid in self.uuids:
            self.uuids.remove(uuid)
        else:
            logger.error("Uuid %s not in uuids %s", client.uuid, self.uuids)

        if self.uuids:
            if self.baton == client.uuid:
                self.baton = self.uuids[0]
            await self.update_baton()
            return True

        self.baton = None
        return False

    async def send_baton_approval_request(self, message: PlotMessage) -> None:
        """Sends message to current baton holder to request baton
        Parameters
        ----------
        message : PlotMessage
        """
        requester = message.params
        if self.baton is None:
            logger.warning("Ignoring baton request as client does not have baton")
        elif requester in self.uuids:
            processed_msg = BatonRequestMessage(requester=requester)
            logger.debug("Baton approved for %s", requester)
            msg = ws_pack(processed_msg)
            if msg is not None:
                for c in self.clients_with_uuid(self.baton):
                    await c.add_message(msg)
            await self.send_next_message()
        else:
            logger.warning("Ignoring baton request by unknown %s", requester)

    async def take_baton(self, message: PlotMessage) -> bool:
        """Updates baton and sends new baton messages
        Parameters
        ----------
        message : PlotMessage

        Returns
        -------
        True if messages updated
        """
        uuid = message.params
        if uuid in self.uuids:
            self.baton = uuid
            await self.update_baton()
            return True

        logger.warning("Ignoring baton request by unknown %s", uuid)
        return False

    def clients_available(self) -> bool:
        """Return True if any clients are available"""
        for cl in self._clients.values():
            if any(cl):
                return True
        return False

    def get_plot_ids(self) -> list[str]:
        """Get plot IDs
        Returns sorted list of plot IDs in all plot clients
        """
        return sorted(list(self._clients.keys()))

    async def clear_plot_states(self, plot_id: str):
        """
        Clears plot_states for a given plot ID

        Parameters
        ----------
        plot_id : str
            ID of plot to clear
        """
        if plot_id in self.plot_states:
            plot_state = self.plot_states[plot_id]
            async with plot_state.lock:
                plot_state.clear()

    async def get_regions(self, plot_id: str) -> list[SelectionBase]:
        """
        Get regions for a given plot ID

        Parameters
        ----------
        plot_id : str

        Returns list of regions from given plot ID
        """
        plot_state = self.plot_states[plot_id]
        async with plot_state.lock:
            cs = plot_state.current_selections
            return [] if cs is None else list(cs)

    async def clear_queues(self, plot_id: str):
        """
        Clears current data, selections and queues for a given plot ID

        Parameters
        ----------
        plot_id : str
            ID of plot to clear
        """
        await self.clear_plot_states(plot_id)
        for c in self._clients[plot_id]:
            c.clear_queue()

    async def clear_plots(self, plot_id: str):
        """
        Sends message to clear plots to clients for a given plot_id

        Parameters
        ----------
        plot_id : str
            ID of plot to which to send data message.
        """
        msg = ws_pack(ClearPlotsMessage(plot_id=plot_id))
        if msg is not None:
            for c in self._clients[plot_id]:
                await c.add_message(msg)
        await self.send_next_message()

    async def clear_plots_and_queues(self, plot_id: str):
        """
        Clears queues and sends message to clear plots to clients
        for a given plot_id

        Parameters
        ----------
        plot_id : str
            ID of plot to which to send data message.
        """
        await self.clear_queues(plot_id)
        await self.clear_plots(plot_id)

    async def benchmark(self, plot_id: str, params: _benchmark.BenchmarkParams) -> str:
        b = getattr(_benchmark, params.plot_type, None)
        if b is None:
            return f"Benchmark type not supported: {params.plot_type}"

        start = -1
        pause = params.pause
        for _ in range(params.iterations):
            for msg in b(params.params):
                msg = ws_pack(msg)
                if msg is None:
                    continue
                for c in self._clients[plot_id]:
                    await c.add_message(msg)
                await self.send_next_message()
                if start == -1:
                    await sleep(1)  # allow more time to setup plot
                    start = time_ns()
                else:
                    await sleep(pause)

        return f"Finished in {int((time_ns() - start)/1000000)}ms"

    async def send_next_message(self):
        """Sends the next response on the response list and updates the client status"""
        if self.clients_available():
            self.client_status = StatusType.busy
            for cl in self._clients.values():
                for c in cl:
                    await c.send_next_message()

    def convert_line_params_to_data_message(
        self, plot_id: str, line_param_msg: ClientLineParametersMessage
    ) -> MultiLineDataMessage:
        """
        Creates new MultiLineDataMessage from existing line data and new parameters

        Parameters
        ----------
        plot_id: str
            id of plot for which to update data
        line_param_msg : ClientLineParametersMessage
            line with updates parameters.
        """

        ml_data_msg = self.plot_states[plot_id].current_data
        if not isinstance(ml_data_msg, MultiLineDataMessage):
            raise ValueError(
                f"Wrong type of message given: MultiLineDataMessage expected: {type(ml_data_msg)}"
            )

        current_lines = ml_data_msg.ml_data
        modified_line_params = line_param_msg.line_params
        updated_lines = []
        key_found = False

        for line in current_lines:
            if not key_found and line.key == line_param_msg.key:
                updated_line = LineData(
                    key=line.key,
                    line_params=modified_line_params,
                    x=line.x,
                    y=line.y,
                    default_indices=line.default_indices,
                )
                updated_lines.append(updated_line)
                key_found = True
            else:
                updated_lines.append(line)

        if not key_found:
            raise ValueError(
                f"No line with key {line_param_msg.key} found in current line data {current_lines}"
            )

        add_colour_to_lines(updated_lines)

        return MultiLineDataMessage(
            ml_data=updated_lines, plot_config=ml_data_msg.plot_config
        )

    def convert_scatter_params_to_data_message(
        self, plot_id: str, scatter_param_msg: ClientScatterParametersMessage
    ) -> ScatterDataMessage:
        """
        Creates new ScatterDataMessage from existing scatter data and new parameters

        Parameters
        ----------
        plot_id: str
            id of plot for which to update data
        scatter_param_msg : ClientScatterParametersMessage
            scatter data parameters.
        """

        sc_data_msg = self.plot_states[plot_id].current_data
        if not isinstance(sc_data_msg, ScatterDataMessage):
            raise ValueError(
                f"Wrong type of message given: ScatterDataMessage expected: {type(sc_data_msg)}"
            )

        sc_data = sc_data_msg.sc_data

        updated_data = ScatterData(
            x=sc_data.x,
            y=sc_data.y,
            point_values=sc_data.point_values,
            domain=sc_data.domain,
            colour_map=sc_data.colour_map,
            point_size=scatter_param_msg.point_size,
        )

        return ScatterDataMessage(
            sc_data=updated_data, plot_config=sc_data_msg.plot_config
        )

    def combine_line_messages(
        self, plot_id: str, new_points_msg: AppendLineDataMessage
    ) -> tuple[MultiLineDataMessage, AppendLineDataMessage]:
        """
        Adds indices to data message and appends points to current multi-line
        data message

        Parameters
        ----------
        plot_id: str
            id of plot to append data to
        new_points_msg : AppendLineDataMessage
            new points to append to current data lines.
        """
        ml_data_msg = self.plot_states[plot_id].current_data
        if not isinstance(ml_data_msg, MultiLineDataMessage):
            raise ValueError(
                f"Wrong type of message given: MultiLineDataMessage expected: {type(ml_data_msg)}"
            )

        current_lines = ml_data_msg.ml_data
        add_colour_to_lines(new_points_msg.al_data)
        new_points = new_points_msg.al_data
        default_indices = current_lines[0].default_indices
        current_lines_len = len(current_lines)
        new_points_len = len(new_points)

        def _append(a: DvDNDArray | None, b: DvDNDArray | None):
            if a is None:
                return b
            if b is None:
                return a
            return np.append(a, b)

        if not default_indices:
            combined_lines = [
                LineData(
                    line_params=c.line_params,
                    x=_append(c.x, p.x),
                    y=np.append(c.y, p.y),
                    default_indices=False,
                )
                for c, p in zip(current_lines, new_points)
            ]

            if current_lines_len > new_points_len:
                combined_lines += current_lines[new_points_len:]

            elif new_points_len > current_lines_len:
                combined_lines += new_points[current_lines_len:]

        else:
            indexed_lines = []
            combined_lines = []
            for c, p in zip(current_lines, new_points):
                c_y_size = c.y.size
                total_y_size = c_y_size + p.y.size
                indexed_lines.append(
                    LineData(
                        line_params=p.line_params,
                        x=np.arange(
                            c_y_size,
                            total_y_size,
                            dtype=np.min_scalar_type(total_y_size),
                        ),
                        y=p.y,
                        default_indices=True,
                    )
                )
                combined_lines.append(
                    LineData(
                        line_params=c.line_params,
                        x=_append(
                            c.x,
                            np.arange(
                                c_y_size,
                                total_y_size,
                                dtype=np.min_scalar_type(total_y_size),
                            ),
                        ),
                        y=np.append(c.y, p.y),
                        default_indices=True,
                    )
                )
            if current_lines_len > new_points_len:
                combined_lines += current_lines[new_points_len:]

            elif new_points_len > current_lines_len:
                extra_indexed_lines = [
                    LineData(
                        line_params=p.line_params,
                        x=np.arange(p.y.size, dtype=np.min_scalar_type(p.y.size)),
                        y=p.y,
                        default_indices=True,
                    )
                    for p in new_points[current_lines_len:]
                ]
                combined_lines += extra_indexed_lines
                indexed_lines += extra_indexed_lines

            new_points_msg.al_data = indexed_lines

        return (
            MultiLineDataMessage(
                ml_data=combined_lines, plot_config=ml_data_msg.plot_config
            ),
            new_points_msg,
        )

    async def update_plot_states_with_message(
        self, msg: DataMessage | SelectionMessage | BatonMessage, plot_id: str
    ) -> bytes | None:
        """Indexes and combines line messages if needed and updates plot states

        Parameters
        ----------
        msg : DataMessage | SelectionMessage
            A message for plot states.
        """
        plot_state = self.plot_states[plot_id]
        new_msg = None
        async with plot_state.lock:
            match msg:
                case BatonMessage():
                    plot_state.current_baton = msg.baton
                    new_msg = plot_state.new_baton_message = ws_pack(msg)

                case SelectionsMessage():
                    plot_state.current_selections = msg.set_selections
                    new_msg = plot_state.new_selections_message = ws_pack(msg)

                case UpdateSelectionsMessage():
                    current = plot_state.current_selections
                    if current is None:
                        current = plot_state.current_selections = []
                    extra = []
                    for u in msg.update_selections:
                        uid = u.id
                        for i, c in enumerate(current):
                            if uid == c.id:
                                current[i] = u
                                break
                        else:
                            extra.append(u)
                    current.extend(extra)
                    plot_state.new_selections_message = ws_pack(
                        SelectionsMessage(set_selections=current)
                    )
                    new_msg = ws_pack(msg)

                case ClientLineParametersMessage():
                    msg = plot_state.current_data = (
                        self.convert_line_params_to_data_message(plot_id, msg)
                    )
                    new_msg = plot_state.new_data_message = ws_pack(msg)

                case ClientScatterParametersMessage():
                    msg = plot_state.current_data = (
                        self.convert_scatter_params_to_data_message(plot_id, msg)
                    )
                    new_msg = plot_state.new_data_message = ws_pack(msg)

                case ClearSelectionsMessage():
                    ids = msg.selection_ids
                    current = plot_state.current_selections
                    if current is None:
                        return
                    if len(ids) == 0:
                        current.clear()
                    else:
                        for s in current:
                            if s.id in ids:
                                current.remove(s)
                    plot_state.new_selections_message = ws_pack(
                        SelectionsMessage(set_selections=current)
                    )
                    new_msg = ws_pack(msg)

                case AppendLineDataMessage():
                    if isinstance(plot_state.current_data, MultiLineDataMessage):
                        combined_msgs, indexed_append_msgs = self.combine_line_messages(
                            plot_id, msg
                        )
                        plot_state.current_data = combined_msgs
                        plot_state.new_data_message = ws_pack(plot_state.current_data)
                        msg = indexed_append_msgs
                    else:
                        msg = convert_append_to_multi_line_data_message(msg)
                    new_msg = ws_pack(msg)

                case DataMessage():

                    def check_cm(
                        cm_id: str, data: HeatmapData | ScatterData | SurfaceData
                    ):
                        if data.colour_map:
                            self.last_colour_maps[cm_id] = data.colour_map
                        else:
                            data.colour_map = self.last_colour_maps[cm_id]

                    if isinstance(msg, MultiLineDataMessage):
                        msg = add_indices(msg)
                        add_colour_to_lines(msg.ml_data)
                    elif isinstance(msg, ImageDataMessage) and isinstance(
                        msg.im_data, HeatmapData
                    ):
                        check_cm("HM:" + plot_id, msg.im_data)
                    elif isinstance(msg, ScatterDataMessage):
                        check_cm("SC:" + plot_id, msg.sc_data)
                    elif isinstance(msg, SurfaceDataMessage):
                        check_cm("SU:" + plot_id, msg.su_data)

                    plot_state.current_data = msg
                    new_msg = plot_state.new_data_message = ws_pack(msg)

        return new_msg

    async def prepare_data(
        self, msg: PlotMessage, omit_client: PlotClient | None = None
    ):
        """Processes PlotMessage into a client message and adds that to any client

        Parameters
        ----------
        msg : PlotMessage
            A client message for processing.
        """
        plot_id = msg.plot_id
        logger.debug("prepare_data %s: %s", type(msg), msg)
        try:
            processed_msg = self.processor.process(msg)
        except Exception:
            logger.error("Could not process message: %s", msg, exc_info=True)
            return

        new_msg = await self.update_plot_states_with_message(processed_msg, plot_id)
        if new_msg is not None:
            for c in self._clients[plot_id]:
                if c is not omit_client:
                    await c.add_message(new_msg)

    def clients_with_uuid(self, uuid: str):
        return (c for cl in self._clients.values() for c in cl if c.uuid == uuid)


async def handle_client(server: PlotServer, plot_id: str, socket: WebSocket, uuid: str):
    client = await server.add_client(plot_id, socket, uuid)
    initialize = True
    try:
        while True:
            update_all = False
            message = await socket.receive()
            if message["type"] == "websocket.disconnect":
                logger.debug("Websocket disconnected: %s:%s", client.name, client.uuid)
                update_all = await server.remove_client(plot_id, client)
                break

            message = ws_unpack(message["bytes"])
            try:
                received_message = PlotMessage.model_validate(message)
            except ValidationError as v_err:
                logger.warn("Ignoring message which may be corrupted: %s", v_err)
                # continue

            if received_message.type == MsgType.status:
                if received_message.params == StatusType.ready:
                    if initialize:
                        await client.send_next_message()
                        await client.send_next_message()  # in case there are selections

                        initialize = False
                    else:
                        server.client_status = StatusType.ready
                    update_all = True
                elif received_message.params == StatusType.closing:
                    logger.info("Websocket closing for %s:%s", client.name, client.uuid)
                    update_all = await server.remove_client(plot_id, client)
                    break

            elif received_message.type == MsgType.baton_request:
                await server.send_baton_approval_request(received_message)

            elif received_message.type == MsgType.baton_offer:
                if uuid == server.baton:
                    update_all = await server.take_baton(received_message)
                else:
                    logger.warning("Baton approval received from non-baton holder")

            else:  # should process events from client (if that client is in control)
                omit = None
                mtype = received_message.type

                is_valid = True
                if (
                    mtype == MsgType.client_new_selection
                    or mtype == MsgType.client_update_selection
                    or mtype == MsgType.clear_selection_data
                    or mtype == MsgType.client_update_line_parameters
                ):
                    logger.debug(
                        "Got from %s (%s) from client %s: %s",
                        plot_id,
                        mtype,
                        client.uuid,
                        received_message.params,
                    )
                    is_valid = client.uuid == server.baton
                    if is_valid:
                        omit = client  # omit originating client
                    else:
                        logger.error(
                            "Selection change requested from client %s without baton",
                            client.uuid,
                        )

                if is_valid:
                    await server.prepare_data(received_message, omit_client=omit)
                    update_all = True

            if update_all:
                await server.send_next_message()

    except WebSocketDisconnect:
        logger.error(
            "Websocket disconnected: %s:%s", client.name, client.uuid, exc_info=True
        )
        update_all = await server.remove_client(plot_id, client)

    if update_all:
        await server.send_next_message()


def add_indices(msg: MultiLineDataMessage) -> MultiLineDataMessage:
    """Adds default indices to a multi-line data message if default indices flag
    is True

    Parameters
    ----------
    msg : MultiLineDataMessage
        A multi-line data message to which to add indices.

    Returns the new multi-line data message
    """
    if msg.ml_data[0].default_indices:
        for m in msg.ml_data:
            m.x = np.arange(m.y.size, dtype=np.min_scalar_type(m.y.size))
    return msg


def convert_append_to_multi_line_data_message(
    msg: AppendLineDataMessage,
) -> MultiLineDataMessage:
    """Converts append line data message to a multi-line data message and adds default
      indices if any default indices flag is True
    indices if any default indices flag is True

    Parameters
    ----------
    msg : AppendLineDataMessage
        An append line data message to convert to a multi-line data message.

    Returns the new multi-line data message
    """
    default_indices = any([a.default_indices for a in msg.al_data])
    if default_indices:
        for m in msg.al_data:
            m.x = np.arange(m.y.size, dtype=np.min_scalar_type(m.y.size))
            m.default_indices = True

    add_colour_to_lines(msg.al_data)

    return MultiLineDataMessage(
        plot_config=msg.plot_config,
        ml_data=msg.al_data,
    )



from time import sleep

import numpy as np
from davidia.models.parameters import Aspect, TableDisplayType
from davidia.plot import (
    LinearSelection,
    RectangularSelection,
    clear,
    image,
    line,
    region,
    scatter,
    surface,
    table,
    set_default_plot_server,
)


def line_demo(p, no_x=False, line_on=False):
    data = [1.5, 4.5, 3.5]
    if no_x:
        x = data
        y = None
        x_label = "index"
    else:
        x = [4.8, 10, 15.1]
        y = data
        x_label = "x-axis"
    line(
        x=x,
        y=y,
        plot_config={
            "x_label": x_label,
            "y_label": "y-axis",
            "x_scale": "linear",
            "y_scale": "linear",
            "title": "line demo plot",
        },
        plot_id=f"plot_{p}",
        line_on=line_on,
        point_size=8,
        glyph_type="Circle",
        colour="red",
    )


def multiline_demo(p, no_x=False, lines=3):
    data = [[(v + 0.2 * i) for v in (1.5, 4.5, 3.5)] for i in range(lines)]
    if no_x:
        x = data
        yds = None
        x_label = "index"
    else:
        x = [4.8, 10, 15.1]
        yds = data
        x_label = "x-axis"
    line(
        x=x,
        y=yds,
        plot_config={
            "x_label": x_label,
            "y_label": "y-axis",
            "x_scale": "linear",
            "y_scale": "linear",
            "title": "multiline demo plot",
        },
        plot_id=f"plot_{p}",
        line_on=True,
    )


def append_lines_demo(p, no_x=False):
    data = [[-2.5, 0, 2.5], [], [14, 15, 14.2]]
    if no_x:
        x = data
        yds = None
    else:
        x = [[20, 25, 30], [], [20, 25, 30]]
        yds = data
    line(x=x, y=yds, append=True, plot_id=f"plot_{p}", line_on=True)


def append_more_lines_demo(p, no_x=False):
    data = [[6, 10, 8], [5, 10, 17.2], []]
    if no_x:
        x = data
        yds = None
    else:
        x = [[35, 40, 45], [20, 25, 30], []]
        yds = data
    line(x=x, y=yds, append=True, plot_id=f"plot_{p}", line_on=True)


def heatmap_demo(p, high=False):
    image(
        [[5, 30, 45], [1.5, 4.5, 3.5]] if high else [[5, 10, 15], [1.5, 4.5, 3.5]],
        domain=[0, 50] if high else [0, 20],
        heatmap_scale="linear",
        colour_map="Inferno",
        aspect=Aspect.auto,
        plot_config={
            "x_label": "x-axis",
            "y_label": "y-axis",
            "x_values": [18, 20, 22, 24],
            "y_values": [-4, 0, 4],
            "title": "heatmap demo plot",
        },
        plot_id=f"plot_{p}",
    )


def image_demo(p):
    image(
        [
            [[0, 255, 255], [255, 0, 255], [255, 255, 0], [0, 0, 255]],
            [[0, 0, 0], [85, 85, 85], [255, 255, 255], [170, 170, 170]],
        ],
        aspect="equal",
        plot_config={
            "x_label": "x-axis",
            "y_label": "y-axis",
            "x_values": [4, 5, 6, 7, 8],
            "y_values": [7, 10, 13],
            "title": "image demo plot",
        },
        plot_id=f"plot_{p}",
    )


def scatter_demo(p):
    scatter(
        x=[max(0.1, x) % 20 for x in range(20)],
        y=[y % 10 for y in range(20)],
        point_values=[6 * i for i in range(20)],
        domain=(0, 114),
        plot_config={
            "x_label": "x-axis",
            "y_label": "y-axis",
            "x_scale": "log",
            "y_scale": "linear",
            "title": "scatter demo plot",
        },
        plot_id=f"plot_{p}",
        colour_map="Cividis",
    )


def surface_demo(p):
    xx, yy = np.meshgrid(np.arange(-3, 6.0), np.array([-2, -0.5, 0, 1, 2.5, 1, 0, -1]))
    surface_data = np.sin(xx) + yy
    surface(
        surface_data,
        domain=(-4, 4),
        surface_scale="linear",
        colour_map="Turbo",
        plot_config={
            "x_label": "x-axis",
            "y_label": "y-axis",
            "x_values": [18, 20, 22, 24, 26, 28, 30, 32],
            "y_values": [-12, -10, -8, -4, 0, 4, 8, 10, 12],
            "title": "surface demo plot",
        },
        plot_id=f"plot_{p}",
    )


def table_demo(p):
    table(
        [[6.23 * i for i in range(20)]] * 5,
        cell_width=120,
        display_style=TableDisplayType.scientific,
        number_digits=1,
        plot_id=f"plot_{p}",
    )


def regions_demo(p):
    region(
        selections=[
            RectangularSelection(
                start=(3.5, 6.5),
                lengths=(3.2, 2.0),
                angle=0.1,
                colour="green",
                alpha=0.7,
            ),
            LinearSelection(
                start=(6.9, 10.5), length=2.0, angle=1.1, colour="cyan", alpha=0.6
            ),
        ],
        plot_id=f"plot_{p}",
    )
    rs = region(plot_id=f"plot_{p}")
    print(rs)
    return rs


def run_line_demos(p, wait=3, no_x=False, verbose=False):
    line_demo(p, no_x)
    if verbose:
        print("line")
    sleep(wait)
    multiline_demo(p, no_x)
    if verbose:
        print("multiline")
    sleep(wait)
    append_lines_demo(p, no_x)
    if verbose:
        print("append line")
    sleep(wait)
    append_more_lines_demo(p, no_x)
    if verbose:
        print("append line 2")
    sleep(wait)


def run_all_demos(wait=3, repeats=5):
    """Run all demos in example app which has plot_0 and plot_1

    Parameters
    ----------
    wait : int, optional
        time to wait between plot changes, by default 3
    repeats : int, optional
        number of times to show all demos, by default 5
    """

    p = 0
    for _ in range(repeats):
        line_demo(p)
        sleep(wait)

        p = 1 - p
        clear(f"plot_{p}")
        line_demo(p, True)
        sleep(wait)

        p = 1 - p
        clear(f"plot_{p}")
        heatmap_demo(p)
        sleep(wait)

        p = 1 - p
        run_line_demos(p, wait)

        p = 1 - p
        run_line_demos(p, wait, True)

        p = 1 - p
        clear(f"plot_{p}")
        image_demo(p)
        regions_demo(p)
        sleep(wait)

        p = 1 - p
        clear(f"plot_{p}")
        scatter_demo(p)
        sleep(wait)

        p = 1 - p
        clear(f"plot_{p}")
        surface_demo(p)
        sleep(wait)

        p = 1 - p
        clear(f"plot_{p}")
        table_demo(p)
        sleep(wait)

        p = 1 - p
        clear(f"plot_{p}")

def create_parser():
    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

    parser = ArgumentParser(
        description="Simple demo of plotting client and server",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-r", "--repetitions", help="Number of repeats", type=int, default=5
    )
    parser.add_argument(
        "-p",
        "--pause",
        help="Period to pause between changes (in seconds)",
        type=float,
        default=3.0,
    )
    parser.add_argument(
        "-H", "--host", help="Set the host address for server", default="127.0.0.1"
    )
    parser.add_argument(
        "-P", "--port", help="Set the port number for server", type=int, default=8000
    )
    return parser

def start_and_run_all_demos():
    from threading import Thread
    from time import sleep
    import webbrowser
    from davidia.main import run_app

    args = create_parser().parse_args()

    def browser():
        sleep(2)
        webbrowser.open(f"http://{args.host}:{args.port}")

    def demo():
        sleep(5)
        set_default_plot_server(args.host, args.port)
        run_all_demos(args.pause, args.repetitions)

    Thread(target=browser).start()
    Thread(target=demo).start()
    run_app(host=args.host, port=args.port)


if __name__ == "__main__":
    args = create_parser().parse_args()
    set_default_plot_server(args.host, args.port)
    run_all_demos(args.pause, args.repetitions)



from __future__ import annotations

from pydantic.alias_generators import to_camel

from ..models.messages import (
    AppendLineDataMessage,
    PlotConfig,
    ClearSelectionsMessage,
    ClientScatterParametersMessage,
    ClientSelectionMessage,
    ClientLineParametersMessage,
    HeatmapData,
    ImageData,
    ImageDataMessage,
    LineData,
    LineParams,
    MsgType,
    MultiLineDataMessage,
    PlotMessage,
    ScatterData,
    ScatterDataMessage,
    SelectionsMessage,
    SurfaceData,
    SurfaceDataMessage,
    TableData,
    TableDataMessage,
    UpdateSelectionsMessage,
)
from ..models.selections import AnySelection, as_selection


def check_line_names(lines: list[LineData]) -> list[LineData]:
    """Autonames lines that do not have names"""
    used_names = set(line.line_params.name for line in lines if line.line_params.name)
    index = 0
    for line in lines:
        if not line.line_params.name:
            new_name = f"Line {index}"
            index += 1
            while new_name in used_names:
                new_name = f"Line {index}"
                index += 1
            line.line_params.name = new_name
            used_names.add(new_name)
    return lines


def extract_selection(raw: dict, entry: str) -> AnySelection | list[AnySelection]:
    if entry not in raw:
        camel_entry = to_camel(entry)
        if camel_entry == entry or camel_entry not in raw:
            raise ValueError(
                f"{entry} and its camelCase {camel_entry} not found in message"
            )
        entry = camel_entry
    raw_selection = raw[entry]
    if isinstance(raw_selection, list):
        return [as_selection(s) for s in raw_selection]
    return as_selection(raw_selection)


class Processor:
    """
    A Processor class used to convert new data request messages to data messages

    ...

    Methods
    -------
    process(message: PlotMessage) -> MultiLineDataMessage | ImageDataMessage
        Converts a PlotMessage to a processed data message
    prepare_new_multiline_data_message(params: list(LineData)
    ) -> MultiLineDataMessage:
        Converts parameters for new lines to a new multiline data messages
    """

    def __init__(self):
        pass

    def process(
        self, message: PlotMessage
    ) -> (
        ClearSelectionsMessage
        | MultiLineDataMessage
        | AppendLineDataMessage
        | ImageDataMessage
        | ScatterDataMessage
        | SelectionsMessage
        | SurfaceDataMessage
        | TableDataMessage
        | UpdateSelectionsMessage
    ):
        """Converts a PlotMessage to processed data

        Parameters
        ----------
        message : PlotMessage
            The message for processing

        Returns
        -------
        ClearSelectionsMessage | MultiLineDataMessage | AppendLineDataMessage
        | ImageDataMessage | ScatterDataMessage | SurfaceDataMessage | TableDataMessage
        | UpdateSelectionsMessage
            The processed data as a message

        Raises
        ------
        ValueError
            If message type is unexpected.
        """

        plot_config = getattr(message, "plot_config", None)
        if plot_config is None:
            plot_config = PlotConfig()
        elif not isinstance(plot_config, PlotConfig):
            plot_config = PlotConfig.model_validate(plot_config)

        params = message.params
        match message.type:
            case MsgType.new_multiline_data | MsgType.append_line_data:
                params = [
                    LineData.model_validate(p) if not isinstance(p, LineData) else p
                    for p in params
                ]
                append = message.type == MsgType.append_line_data
                return self.prepare_new_multiline_data_message(
                    params, plot_config, append=append
                )
            case MsgType.new_image_data:
                if not isinstance(params, ImageData):
                    if "domain" in params:
                        params = HeatmapData.model_validate(params)
                    else:
                        params = ImageData.model_validate(params)
                return ImageDataMessage(im_data=params, plot_config=plot_config)
            case MsgType.new_scatter_data:
                if not isinstance(params, ScatterData):
                    params = ScatterData.model_validate(params)
                return ScatterDataMessage(sc_data=params, plot_config=plot_config)
            case MsgType.new_surface_data:
                if not isinstance(params, SurfaceData):
                    params = SurfaceData.model_validate(params)
                return SurfaceDataMessage(su_data=params, plot_config=plot_config)
            case MsgType.new_table_data:
                if not isinstance(params, TableData):
                    params = TableData.model_validate(params)
                return TableDataMessage(ta_data=params, plot_config=plot_config)
            case MsgType.client_new_selection | MsgType.client_update_selection:
                if not isinstance(params, ClientSelectionMessage):
                    params = ClientSelectionMessage(
                        selection=extract_selection(params, "selection")
                    )
                return UpdateSelectionsMessage(update_selections=[params.selection])
            case MsgType.client_update_line_parameters:
                if not isinstance(params, ClientLineParametersMessage):
                    key = params["key"]
                    line = params["line_params"]
                    line_params = (
                        LineParams.model_validate(line)
                        if not isinstance(line, LineParams)
                        else line
                    )
                    params = ClientLineParametersMessage(
                        key=key, line_params=line_params
                    )
                return params
            case MsgType.client_update_scatter_parameters:
                if not isinstance(params, ClientScatterParametersMessage):
                    params = ClientScatterParametersMessage(
                        point_size=params["point_size"]
                    )
                return params
            case MsgType.new_selection_data:
                if not isinstance(params, SelectionsMessage):
                    params = SelectionsMessage(
                        set_selections=extract_selection(params, "set_selections")
                    )
                return params
            case MsgType.update_selection_data:
                if not isinstance(params, UpdateSelectionsMessage):
                    params = UpdateSelectionsMessage(
                        update_selections=extract_selection(params, "update_selections")
                    )
                return params
            case MsgType.clear_selection_data:
                if not isinstance(params, ClearSelectionsMessage):
                    params = ClearSelectionsMessage.model_validate(params)
                return params
            case _:
                # not covered by tests
                raise ValueError(f"message type not in list: {message.type}")

    def prepare_new_multiline_data_message(
        self, params: list[LineData], plot_config: PlotConfig, append=False
    ) -> MultiLineDataMessage | AppendLineDataMessage:
        """Converts parameters for a new line to processed new line data

        Parameters
        ----------
        params : list[LineData]
            List of line data parameters to be processed to new multiline data
        plot_config : PlotConfig
            Plot configuration parameters
        append : returns AppendLineDataMessage

        Returns
        -------
        MultiLineDataMessage | AppendLineDataMessage
            New multiline data message.
        """

        params = check_line_names(params)

        if append:
            return AppendLineDataMessage(al_data=params, plot_config=plot_config)
        return MultiLineDataMessage(ml_data=params, plot_config=plot_config)



from __future__ import annotations

import logging
import os
import pathlib

import uvicorn
from davidia.models.messages import PlotMessage
from davidia.models.selections import AnySelection
from davidia.server.benchmarks import BenchmarkParams
from davidia.server.fastapi_utils import message_unpack
from davidia.server.plotserver import PlotServer, handle_client
from fastapi import FastAPI, WebSocket
from fastapi.middleware.cors import CORSMiddleware  # comment this on deployment
from fastapi.openapi.utils import get_openapi
from fastapi.staticfiles import StaticFiles

_PM_SCHEMA = PlotMessage.model_json_schema(
    ref_template="#/components/schemas/{model}", by_alias=True
)
_PM_NESTED_MODELS = _PM_SCHEMA.pop("$defs")

logger = logging.getLogger("main")


def _create_bare_app(add_benchmark=False):
    app = FastAPI()

    def customize_openapi():
        if app.openapi_schema:
            return app.openapi_schema

        o_schema = get_openapi(title="Davidia API", version="1.0.0", routes=app.routes)
        o_schema["components"]["schemas"].update(_PM_NESTED_MODELS)
        app.openapi_schema = o_schema
        return o_schema

    app.openapi = customize_openapi
    origins = ["*"]
    app.add_middleware(
        CORSMiddleware, allow_origins=origins
    )  # comment this on deployment
    ps = PlotServer()
    setattr(app, "_plot_server", ps)

    @app.websocket("/plot/{uuid}/{plot_id}")
    async def websocket(websocket: WebSocket, uuid: str, plot_id: str):
        """End point for plot server to web UI communication.

        PlotMessages are passed between client/server
        """
        await websocket.accept()
        await handle_client(ps, plot_id, websocket, uuid)

    @app.post(
        "/push_data",
        openapi_extra={
            "requestBody": {
                "content": {
                    "application/x-yaml": {"schema": _PM_SCHEMA},
                },
                "required": True,
            }
        },
    )
    @message_unpack
    async def push_data(data: PlotMessage) -> str:
        """
        Push data to plot
        """
        await ps.prepare_data(data)
        await ps.send_next_message()
        return "data sent"

    @app.put("/clear_data/{plot_id}")
    async def clear_data(plot_id: str) -> str:
        """
        Clear plot
        """
        await ps.clear_plots_and_queues(plot_id)
        return "data cleared"

    @app.get("/get_plot_ids")
    def get_plot_ids() -> list[str]:
        return ps.get_plot_ids()

    @app.get("/get_regions/{plot_id}")
    async def get_regions(plot_id: str) -> list[AnySelection]:
        return await ps.get_regions(plot_id)

    if add_benchmark:

        @app.post("/benchmark/{plot_id}")
        async def benchmark(plot_id: str, params: BenchmarkParams) -> str:
            """
            Benchmark davidia
            """
            return await ps.benchmark(plot_id, params)

    return app


def _setup_logger():
    ch = logging.StreamHandler()
    ch.setFormatter(
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )
    logger.addHandler(ch)
    logger.setLevel(logging.DEBUG)


def add_client_endpoint(app, client_path):
    index_path = client_path / "index.html"
    if index_path.is_file() and os.access(index_path, os.R_OK):
        logger.debug("Adding /client endpoint which uses %s", client_path)
        app.mount("/", StaticFiles(directory=client_path, html=True), name="webui")
    else:
        logger.warning(
            "%s not readable so '/' endpoint will not be available", index_path
        )


def find_client_build():
    dvd_pkg_dir = pathlib.Path(__file__).parent
    # example client packaged
    client_dir = dvd_pkg_dir / "client"
    if client_dir.is_dir():
        return client_dir
    # or developing in source tree
    client_dir = dvd_pkg_dir.parent / "example-client/sdist"
    if client_dir.is_dir():
        return client_dir
    logger.warning(
        "Client directory not found in %s (if developing, then build example app)",
        dvd_pkg_dir,
    )


CLIENT_BUILD_PATH = find_client_build()


def create_parser():
    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, SUPPRESS

    CLIENT_BUILD_DIR = str(CLIENT_BUILD_PATH)
    parser = ArgumentParser(
        description="Davidia plot server", formatter_class=ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "-b", "--benchmark", help="Add /benchmark endpoint", action="store_true"
    )
    parser.add_argument(
        "-c",
        "--client",
        help=f"Add /client endpoint using given build directory (or '{CLIENT_BUILD_DIR}')",
        nargs="?",
        const=CLIENT_BUILD_DIR,
        default=SUPPRESS,
    )
    parser.add_argument(
        "-H", "--host", help="Set the host address for server", default="127.0.0.1"
    )
    parser.add_argument(
        "-P", "--port", help="Set the port number for server", type=int, default=80
    )
    return parser


def create_app(client_path=CLIENT_BUILD_PATH, benchmark=False):
    _setup_logger()
    app = _create_bare_app(
        benchmark or os.getenv("DVD_BENCHMARK", "off").lower() == "on"
    )
    if client_path:
        if client_path.is_dir():
            add_client_endpoint(app, client_path)
    return app


def run_app(
    client_path=CLIENT_BUILD_PATH, benchmark=False, host="127.0.0.1", port=80
):
    app = create_app(client_path=client_path, benchmark=benchmark)
    uvicorn.run(app, host=host, port=port, log_level="info", access_log=False)


def main():
    args = create_parser().parse_args()
    client_path = getattr(args, "client", None)
    run_app(
        client_path=pathlib.Path(client_path) if client_path else None,
        benchmark=args.benchmark,
        host=args.host,
        port=args.port,
    )


if __name__ == "__main__":
    main()



from enum import auto
from typing import Any
from uuid import uuid4

from numpy import asanyarray as _asanyarray
from pydantic import ConfigDict, Field, field_validator, model_validator

from .parameters import (
    Aspect,
    AutoNameEnum,
    PlotConfig,
    DvDModel,
    DvDNDArray,
    DvDNpModel,
    ScaleType,
    TableDisplayParams,
    validate_scale_type,
)
from .selections import AnySelection, Float, FloatTuple


class MsgType(AutoNameEnum):
    """Class for message type."""

    status = auto()
    baton_request = auto()
    baton_offer = auto()
    new_multiline_data = auto()
    append_line_data = auto()
    new_image_data = auto()
    new_scatter_data = auto()
    new_surface_data = auto()
    new_table_data = auto()
    new_selection_data = auto()
    update_selection_data = auto()
    clear_selection_data = auto()
    clear_data = auto()
    client_new_selection = auto()
    client_update_selection = auto()
    client_update_line_parameters = auto()
    client_update_scatter_parameters = auto()


class StatusType(AutoNameEnum):
    """Class for status type."""

    ready = auto()
    busy = auto()
    closing = auto()


class GlyphType(AutoNameEnum):
    """Class for glyph type."""

    Circle = auto()
    Cross = auto()
    Square = auto()
    Cap = auto()


class ColourMap(AutoNameEnum):
    Blues = auto()
    Greens = auto()
    Greys = auto()
    Oranges = auto()
    Purples = auto()
    Reds = auto()
    Turbo = auto()
    Viridis = auto()
    Inferno = auto()
    Magma = auto()
    Plasma = auto()
    Cividis = auto()
    Warm = auto()
    Cool = auto()
    Cubehelix = auto()
    BuGn = auto()
    BuPu = auto()
    GnBu = auto()
    OrRd = auto()
    PuBuGn = auto()
    PuBu = auto()
    PuRd = auto()
    RdPu = auto()
    YlGnBu = auto()
    YlGn = auto()
    YlOrBr = auto()
    YlOrRd = auto()
    Rainbow = auto()
    Sinebow = auto()
    HSL = auto()
    BrBG = auto()
    PRGn = auto()
    PiYG = auto()
    PuOr = auto()
    RdBu = auto()
    RdGy = auto()
    RdYlBu = auto()
    RdYlGn = auto()
    Spectral = auto()


class LineParams(DvDModel):
    """Class for representing a line."""

    model_config = ConfigDict(
        extra="forbid",
    )  # need this to prevent any dict validating as all fields have default values

    name: str = ""
    colour: str | None = None
    line_on: bool = True
    point_size: int | None = None
    glyph_type: GlyphType = GlyphType.Circle

    @field_validator("glyph_type")
    @classmethod
    def validate_glyph_type(cls, v: GlyphType | str):
        if isinstance(v, str):
            v = GlyphType[v]
        return v

    @model_validator(mode="before")
    @classmethod
    def check_glyph_type(cls, values: dict):
        if "glyphType" not in values and values.get("glyph_type") is None:
            values["glyph_type"] = GlyphType.Circle
        return values


class LineData(DvDNpModel):
    """Class for representing a line."""

    @staticmethod
    def get_default_key() -> str:
        return uuid4().hex[-8:]

    key: str = Field(default_factory=get_default_key)
    x: DvDNDArray | None = None
    y: DvDNDArray
    line_params: LineParams
    default_indices: bool | None = None

    @field_validator("y")
    @classmethod
    def equal_axes(cls, v, values):
        x_size = getattr(values.data.get("x", 0), "size", 0)
        if x_size != 0:
            y_size = getattr(v, "size", 0)
            if x_size != y_size and x_size != y_size + 1:
                raise ValueError(
                    "x and y arrays must be equal length if provided: "
                    f"{x_size}, {y_size}"
                )
        return v

    @model_validator(mode="before")
    @classmethod
    def are_indices_default(cls, values: Any):
        if not isinstance(values, dict):
            return values

        for k in ("x", "y"):
            if k in values:
                v = values[k]
                if v is None:
                    values.pop(k)
                else:
                    values[k] = _asanyarray(v)

        if not values.get("default_indices"):
            values["default_indices"] = (
                "y" not in values or "x" not in values or values["x"].size == 0
            )
        return values


class ImageData(DvDNpModel):
    """Class for representing an image."""

    values: DvDNDArray
    aspect: Aspect | float | int | None = None
    model_config = ConfigDict(
        extra="forbid"
    )  # need this to prevent any dict validating as all fields have default values


def validate_colour_map(v: ColourMap | str):
    if isinstance(v, str):
        return ColourMap[v]
    return v


class HeatmapData(ImageData):
    """Class for representing heatmap data."""

    domain: FloatTuple
    heatmap_scale: ScaleType = ScaleType.linear
    colour_map: ColourMap | None = None
    "An optional parameter which if None uses the last colour map defined in the plot server which is initialised to a default colour map"

    validate_scale = field_validator("heatmap_scale")(validate_scale_type)
    validate_local_colour_map = field_validator("colour_map")(validate_colour_map)

    @field_validator("colour_map")
    @classmethod
    def validate_colour_map(cls, v: ColourMap | str):
        if isinstance(v, str):
            return ColourMap[v]
        return v


class ScatterData(DvDNpModel):
    """Class for representing scatter data."""

    x: DvDNDArray
    y: DvDNDArray
    point_values: DvDNDArray
    domain: FloatTuple
    colour_map: ColourMap | None = None
    point_size: Float = 10

    validate_local_colour_map = field_validator("colour_map")(validate_colour_map)


class SurfaceData(DvDNpModel):
    """Class for representing surface data."""

    height_values: DvDNDArray
    domain: FloatTuple
    colour_map: ColourMap | None = None
    surface_scale: ScaleType = ScaleType.linear

    validate_scale = field_validator("surface_scale")(validate_scale_type)
    validate_local_colour_map = field_validator("colour_map")(validate_colour_map)


class TableData(DvDNpModel):
    """Class for representing table data."""

    cell_values: DvDNDArray
    cell_width: int
    display_params: TableDisplayParams | None = None


class BatonMessage(DvDModel):
    """Class for representing a baton message."""

    baton: str | None
    uuids: list[str]


class BatonRequestMessage(DvDModel):
    """Class for representing a baton request message."""

    requester: str


class DataMessage(DvDNpModel):
    """Class for representing a data message

    Make sure subclasses have unique fields so client can distinguish
    message type
    """

    plot_config: PlotConfig = PlotConfig()


class AppendLineDataMessage(DataMessage):
    """Class for representing an append line data message."""

    al_data: list[LineData]


class MultiLineDataMessage(DataMessage):
    """Class for representing a multiline data message."""

    ml_data: list[LineData]

    @field_validator("ml_data")
    @classmethod
    def ml_data_is_not_empty(cls, v):
        if len(v) > 0:
            return v
        raise ValueError("ml_data contains no LineData", v)

    @field_validator("ml_data")
    @classmethod
    def default_indices_match(cls, v):
        default_indices = [ld.default_indices for ld in v]
        if all(default_indices) or not any(default_indices):
            return v
        raise ValueError(
            "default_indices must be all True or all False in every LineData object", v
        )


class ImageDataMessage(DataMessage):
    """Class for representing an image data message."""

    im_data: ImageData | HeatmapData


class ScatterDataMessage(DataMessage):
    """Class for representing a scatter data message."""

    sc_data: ScatterData


class SurfaceDataMessage(DataMessage):
    """Class for representing a surface data message."""

    su_data: SurfaceData


class TableDataMessage(DataMessage):
    """Class for representing a table data message."""

    ta_data: TableData


class ClearPlotsMessage(DvDModel):
    """Class for representing a request to clear all plots."""

    plot_id: str


class SelectionMessage(DvDModel):
    """Class to mark selections"""


class ClientSelectionMessage(SelectionMessage):
    """Class for representing a client selection"""

    selection: AnySelection


class ClientLineParametersMessage(DvDModel):
    """Class for representing a client selection"""

    line_params: LineParams
    key: str


class ClientScatterParametersMessage(DvDModel):
    """Class for representing client scatter parameters"""

    point_size: Float


class SelectionsMessage(SelectionMessage):
    """Class for representing a request to set selections"""

    set_selections: list[AnySelection]


class UpdateSelectionsMessage(SelectionMessage):
    """Class for representing a request to update selections"""

    update_selections: list[AnySelection]


class ClearSelectionsMessage(SelectionMessage):
    """Class for representing a request to clear listed or all selections."""

    selection_ids: list[str]


ANY_PM_PARAMS = (
    str
    | list[LineData]
    | HeatmapData
    | ImageData
    | ScatterData
    | SurfaceData
    | TableData
    | ClientSelectionMessage
    | ClientLineParametersMessage
    | ClientScatterParametersMessage
    | SelectionsMessage
    | UpdateSelectionsMessage
    | ClearSelectionsMessage
)


class PlotMessage(DvDNpModel):
    """
    Class for communication messages to push_data endpoint

    Attributes
    ----------
    plot_id : str
        ID of plot to which to send data message
    type : MsgType
        The message type represented as a MsgType enum
    params : Any
        The message params
    plot_config: PlotConfig
        The plot configuration parameters
    """

    plot_id: str
    type: MsgType
    params: ANY_PM_PARAMS = Field(union_mode="left_to_right")
    plot_config: PlotConfig | None = None


ALL_MESSAGES = (
    ClearPlotsMessage,
    AppendLineDataMessage,
    MultiLineDataMessage,
    ImageDataMessage,
    ScatterDataMessage,
    TableDataMessage,
    ClientSelectionMessage,
    ClientLineParametersMessage,
    ClientScatterParametersMessage,
    SelectionsMessage,
    UpdateSelectionsMessage,
    ClearSelectionsMessage,
    BatonMessage,
    BatonRequestMessage,
)

ALL_MODELS = (
    PlotMessage,
    LineData,
    LineParams,
    HeatmapData,
    ImageData,
    ScatterData,
    SurfaceData,
    TableData,
    PlotConfig,
) + ALL_MESSAGES

if __name__ == "__main__":
    import json

    print(json.dumps(PlotMessage.model_json_schema(), indent=2))



from enum import auto as _auto, Enum
from typing import Any, Annotated

import numpy as np
from pydantic import BaseModel, ConfigDict, field_validator
from pydantic.alias_generators import to_camel
from pydantic_numpy.helper.annotation import NpArrayPydanticAnnotation
from pydantic_numpy.model import NumpyModel

DvDNDArray = Annotated[
    np.ndarray[Any, None],
    NpArrayPydanticAnnotation.factory(
        data_type=None,
        dimensions=None,
        strict_data_typing=None,
    ),
]


class AutoNameEnum(str, Enum):
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return name


class Aspect(AutoNameEnum):
    """Class for aspect type."""

    auto = _auto()
    equal = _auto()


class ScaleType(AutoNameEnum):
    """Class for scale type"""

    linear = _auto()
    log = _auto()
    symlog = _auto()
    sqrt = _auto()
    gamma = _auto()


class TableDisplayType(AutoNameEnum):
    """
    Class for table display type
    standard: plain number formatting with decimal places
    scientific: order-of-magnitude.
    """

    standard = _auto()
    scientific = _auto()


class DvDModel(BaseModel):
    """Base model can serialize to and validate camelCase aliases of its attributes"""

    model_config = ConfigDict(
        alias_generator=to_camel, populate_by_name=True, use_enum_values=True
    )


class DvDNpModel(NumpyModel):
    """Base model can serialize to and validate camelCase aliases of its attributes"""

    model_config = ConfigDict(
        alias_generator=to_camel, populate_by_name=True, use_enum_values=True
    )


class TableDisplayParams(DvDModel):
    """
    Class for representing table display type and number of digits

    Attributes
    ----------
    display_type : TableDisplayType | None
        Type of notation for table display
    number_digits : int | None
        The number of significant figures if scientific notation (between 1 and 21)
         or the number of decimal places if standard notation (between 0 and 20).
    """

    model_config = ConfigDict(
        extra="forbid"
    )  # need this to prevent any dict validating as all fields have default values

    display_type: TableDisplayType | None = None
    number_digits: int | None = None

    @field_validator("display_type")
    @classmethod
    def validate_display_type(cls, v: TableDisplayType | str):
        if isinstance(v, str):
            v = TableDisplayType[v]
        return v

    @field_validator("number_digits")
    @classmethod
    def validate_number_digits(cls, v, info):
        if not v:
            return v
        if info.data["display_type"] == TableDisplayType.scientific:
            x = min(v, 21)
            x = max(x, 1)
            return x
        x = min(v, 20)
        x = max(x, 0)
        return x


def validate_scale_type(v: ScaleType | str) -> ScaleType:
    if isinstance(v, str):
        v = ScaleType[v]
    return v


class PlotConfig(DvDNpModel):
    """Class for representing plot configuration."""

    model_config = ConfigDict(
        extra="forbid"
    )  # need this to prevent any dict validating as all fields have default values

    x_label: str = ""
    x_scale: ScaleType = ScaleType.linear
    x_values: DvDNDArray | None = None
    y_label: str = ""
    y_scale: ScaleType = ScaleType.linear
    y_values: DvDNDArray | None = None
    title: str = ""

    validate_x_scale = field_validator("x_scale")(validate_scale_type)
    validate_y_scale = field_validator("y_scale")(validate_scale_type)



import inspect
import logging
from typing import Any

import numpy as np
import orjson
from fastapi import Request, Response
from msgpack import packb as _mp_packb  # max_buffer_size=100MB
from msgpack import unpackb as _mp_unpackb
from pydantic import BaseModel, ValidationError

from ..models.messages import ALL_MODELS, DvDNDArray
from ..models.selections import as_selection

logger = logging.getLogger("main")


def as_model(raw: dict) -> BaseModel | None:
    for m in ALL_MODELS:
        try:
            return m.model_validate(raw)
        except ValidationError:
            logger.debug("Maybe not %s", m)
        except Exception:
            pass
    return None


def j_dumps(data, default=None) -> bytes:
    if isinstance(data, BaseModel):
        return data.model_dump_json().encode()

    d = orjson.dumps(
        data,
        default=default,
        option=orjson.OPT_SERIALIZE_NUMPY,
    )
    return d


def _deserialize_selection(item):
    if isinstance(item, (tuple, list)):
        return [_deserialize_selection(i) for i in item]
    elif isinstance(item, dict):
        if "id" in item and "start" in item:
            return as_selection(item)
    return item


def _deserialize_any(item):
    if isinstance(item, dict):
        m = as_model(item)
        return _deserialize_selection(item) if m is None else m
    elif isinstance(item, list):
        return [_deserialize_any(i) for i in item]
    return item


def j_loads(data):
    d = orjson.loads(data)
    return _deserialize_any(d)


def decode_ndarray(obj) -> DvDNDArray:
    if isinstance(obj, dict):
        if all(i in obj for i in ("nd", "dtype", "shape", "data")) and obj["nd"]:
            obj = np.ndarray(buffer=obj["data"], shape=obj["shape"], dtype=obj["dtype"])
    return obj  # pyright: ignore[reportGeneralTypeIssues]


def encode_ndarray(obj) -> dict[str, Any]:
    if isinstance(obj, np.ndarray):
        kind = obj.dtype.kind
        if kind == "i":  # reduce integer array byte size if possible
            vmin = obj.min() if obj.size > 0 else 0
            if vmin >= 0:
                kind = "u"
            else:
                vmax = obj.max() if obj.size > 0 else 0
                minmax_type = [np.min_scalar_type(vmin), np.min_scalar_type(vmax)]
                if minmax_type[1].kind == "u":
                    isize = minmax_type[1].itemsize
                    stype = np.dtype(f"i{isize}")
                    if isize == 8 and vmax > np.iinfo(stype).max:
                        minmax_type[1] = np.dtype(np.float64)
                    else:
                        minmax_type[1] = stype
                obj = obj.astype(np.promote_types(*minmax_type))
        if kind == "u":
            obj = obj.astype(np.min_scalar_type(obj.max() if obj.size > 0 else 0))
        obj = dict(
            nd=True, dtype=obj.dtype.str, shape=obj.shape, data=obj.data.tobytes()
        )
    return obj


def ws_pack(obj) -> bytes | None:
    """Pack object for a websocket message

    Packs object by converting Pydantic models and ndarrays to dicts before
    using MessagePack
    """
    if isinstance(obj, BaseModel):
        obj = obj.model_dump(by_alias=True)
    return _mp_packb(obj, use_bin_type=True, default=encode_ndarray)


def ws_unpack(obj: bytes) -> dict[str, Any]:
    """Unpack a websocket message as a dict

    Unpacks MessagePack object to dict (deserializes NumPy ndarrays)
    """
    return _mp_unpackb(obj, raw=False, object_hook=decode_ndarray)


_MESSAGE_PACK = "application/x-msgpack"


def message_unpack(func):
    """
    Use with router function:
    @app.get('/')
    @message_unpack
    async def root_request(payload: MyModel) -> OtherModel:
        ...
    """
    f_params = inspect.get_annotations(func, eval_str=True)
    f_class = f_params.pop("return")

    def _instantiate_obj(model_class, obj):
        if isinstance(obj, BaseModel):
            return obj
        elif hasattr(model_class, "model_validate"):
            try:
                return model_class.model_validate(obj)
            except ValidationError as v_err:
                logger.warning("Could not validate: %s", v_err)
                return None
        return model_class(**obj)

    async def wrapper(request: Request) -> Response:
        ct = request.headers.get("Content-Type")
        unpacker = ws_unpack if ct == _MESSAGE_PACK else j_loads
        body = await request.body()
        unpacked = unpacker(body)
        if len(f_params) == 1:
            kwargs = {k: _instantiate_obj(v, unpacked) for k, v in f_params.items()}
        else:
            kwargs = {
                # TODO something about missing parameters or extra items in unpacked
                k: _instantiate_obj(
                    v,
                    unpacked[k],  # pyright: ignore[reportGeneralTypeIssues]
                )
                for k, v in f_params.items()
                if k in unpacked  # pyright: ignore[reportGeneralTypeIssues]
            }

        response = await func(**kwargs)
        if not isinstance(response, f_class):
            raise ValueError(
                f"Return value was not expected type {type(response)} cf {f_class}"
            )

        ac = request.headers.get("Accept")
        packer = ws_pack if ac == _MESSAGE_PACK else j_dumps
        if isinstance(response, Response):
            packed = packer(response.body)
            if packed is not None:
                response.body = packed
        else:
            if isinstance(response, BaseModel):
                response = response.model_dump(by_alias=True)
            response = Response(content=packer(response), media_type=ac)
        return response

    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    return wrapper



import pathlib
from setuptools import setup

__version__ = "1.0.3"


def find_readme():
    current_dir = pathlib.Path(__file__).absolute().parent
    readme = current_dir / "README.md"
    if not readme.is_file():
        import shutil

        parent_dir = current_dir.parent
        shutil.copy(parent_dir / "README.md", current_dir)
    if readme.is_file():
        return readme
    raise FileNotFoundError(f"README.md not found in {current_dir} or {parent_dir}")


readme_path = find_readme()


setup(
    name="davidia",
    version=__version__,
    description="Plot server with React frontend data visualiser",
    long_description=open(readme_path).read(),
    long_description_content_type="text/markdown",
    author_email="dataanalysis@diamond.ac.uk",
    package_dir={
        "davidia": "davidia",
        "davidia.demos": "demos",
    },
    package_data={
        "davidia.data": ["*.png"],
    },
    entry_points={
        "console_scripts": [
            "dvd-server = davidia.main:main",
            "dvd-benchmark = davidia.demos.benchmark:main",
        ],
    },
    python_requires=">=3.10",
    install_requires=[
        "before-after",
        "fastapi",
        "httpx",
        "msgpack",
        "pillow",
        "pydantic-numpy",
        "orjson",
        "requests",
        "uvicorn",
        "websockets",
    ],
    extras_require={
        "test": [
            "pytest",
            "pytest-asyncio",
        ],
        "all": ["davidia-example-client"],
    },
    url="https://github.com/DiamondLightSource/davidia",
)



from __future__ import annotations

import datetime
import itertools
import time
from enum import Enum
from typing import Any, Callable

import numpy as np
import pytest
from dataclasses import asdict, dataclass, is_dataclass
from davidia.main import _create_bare_app
from davidia.models.messages import (
    DvDNDArray,
    LineData,
    LineParams,
    MsgType,
    MultiLineDataMessage,
    PlotMessage,
    StatusType,
)
from davidia.server.fastapi_utils import (
    j_dumps,
    j_loads,
    message_unpack,
    ws_pack,
    ws_unpack,
)
from fastapi.testclient import TestClient
from httpx import AsyncClient
from pydantic import BaseModel
from pydantic_numpy.model import NumpyModel


def test_status_ws():
    data_0 = [
        LineData(
            key="line_0",
            line_params=LineParams(
                colour="red",
                point_size=8,
            ),
            x=[0, 1, 2, 3, 4],
            y=[0, 1, 4, 9, 16],
        ),
        LineData(
            key="line_1",
            line_params=LineParams(colour="blue"),
            x=[2, 4, 6, 8, 9],
            y=[20, 10, 30, 50],
        ),
        LineData(
            key="line_2",
            line_params=LineParams(
                colour="green",
                line_on=False,
                point_size=8,
            ),
            x=[0, 1, 2, 3, 4],
            y=[0, 10, 40, 10, 0],
        ),
    ]
    plot_msg_0 = PlotMessage(
        plot_id="plot_0", type=MsgType.new_multiline_data, params=data_0
    )
    msg_0 = plot_msg_0.model_dump(by_alias=True)
    data_1 = [
        LineData(
            key="line_0",
            line_params=LineParams(
                colour="black",
                point_size=8,
            ),
            x=[0, 1, 2, 3, 4, 5],
            y=[4, 8, 12, 16, 20],
        ),
        LineData(
            key="line_1",
            line_params=LineParams(colour="pink"),
            x=[3, 5, 7, 9, 11],
            y=[-1, -5, 5, 10],
        ),
        LineData(
            key="line_2",
            line_params=LineParams(
                colour="purple",
                line_on=False,
                point_size=8,
            ),
            x=[0, 1, 2, 3, 4],
            y=[0, 20, 30, 10, 10],
        ),
    ]
    plot_msg_1 = PlotMessage(
        plot_id="plot_1", type=MsgType.new_multiline_data, params=data_1
    )
    msg_1 = plot_msg_1.model_dump(by_alias=True)

    data_2 = LineData(
        key="new_line",
        line_params=LineParams(colour="black"),
        x=[10, 20, 30],
        y=[-3, -1, 5],
    )
    plot_msg_2 = PlotMessage(
        plot_id="plot_0", type=MsgType.new_multiline_data, params=[data_2]
    )
    msg_2 = plot_msg_2.model_dump(by_alias=True)

    app = _create_bare_app()

    with TestClient(app) as client:
        with client.websocket_connect("/plot/30064551/plot_0") as ws_0:
            with client.websocket_connect("/plot/30064551/plot_1") as ws_1:
                ps = getattr(app, "_plot_server")
                client_0 = ps._clients["plot_0"]
                client_1 = ps._clients["plot_1"]
                plot_state_0 = ps.plot_states["plot_0"]
                plot_state_1 = ps.plot_states["plot_1"]
                assert ps.client_status == StatusType.busy
                assert len(client_0) == 1
                assert len(client_1) == 1

                assert plot_state_0.current_data is None
                assert plot_state_0.current_selections is None
                assert plot_state_0.new_data_message is None
                assert plot_state_0.new_selections_message is None
                assert plot_state_1.new_data_message is None
                assert plot_state_1.new_selections_message is None
                assert plot_state_1.current_data is None
                assert plot_state_1.current_selections is None

                ws_0.send_bytes(ws_pack(msg_0))
                time.sleep(1)
                assert ps.client_status == StatusType.busy
                assert plot_state_0.current_data
                assert plot_state_0.current_selections is None
                assert plot_state_0.new_data_message
                assert plot_state_0.new_selections_message is None
                assert plot_state_1.new_data_message is None
                assert plot_state_1.new_selections_message is None
                assert plot_state_1.current_data is None
                assert plot_state_1.current_selections is None

                ws_1.send_bytes(ws_pack(msg_1))
                time.sleep(1)
                assert ps.client_status == StatusType.busy
                assert len(client_0) == 1
                assert len(client_1) == 1
                assert plot_state_0.new_data_message
                assert plot_state_1.new_data_message
                assert plot_state_0.new_selections_message is None
                assert plot_state_1.new_selections_message is None
                ws_0.send_bytes(
                    ws_pack(
                        {
                            "plot_id": "plot_0",
                            "type": "status",
                            "params": "ready",
                        }
                    )
                )
                time.sleep(1)
                assert ps.client_status == StatusType.busy

                received_0_0 = ws_0.receive()
                received_0_1 = ws_0.receive()
                rec_text_0_0 = ws_unpack(received_0_0["bytes"])
                rec_text_0_1 = ws_unpack(received_0_1["bytes"])

                assert (
                    rec_text_0_0
                    == rec_text_0_1
                    == {"baton": "30064551", "uuids": ["30064551"]}
                )

                received_0_2 = ws_0.receive()
                rec_text_0_2 = ws_unpack(received_0_2["bytes"])
                nppd_assert_equal(
                    rec_text_0_2["mlData"][2]["y"], np.array([0, 10, 40, 10, 0])
                )

                ws_1.send_bytes(
                    ws_pack(
                        {
                            "plot_id": "plot_1",
                            "type": "status",
                            "params": "ready",
                        }
                    )
                )
                time.sleep(1)
                assert ps.client_status == StatusType.busy

                received_1_0 = ws_1.receive()
                rec_text_1_0 = ws_unpack(received_1_0["bytes"])
                assert rec_text_1_0 == {"baton": "30064551", "uuids": ["30064551"]}

                received_1_1 = ws_1.receive()
                rec_text_1_1 = ws_unpack(received_1_1["bytes"])
                nppd_assert_equal(
                    rec_text_1_1["mlData"][1]["x"], np.array([3, 5, 7, 9, 11])
                )

                ws_0.send_bytes(ws_pack(msg_2))
                time.sleep(1)
                assert ps.client_status == StatusType.busy
                received_new_line = ws_0.receive()
                rec_data = ws_unpack(received_new_line["bytes"])
                line_msg = MultiLineDataMessage.model_validate(rec_data)
                assert line_msg is not None


@dataclass
class Codec:
    mime_type: str
    encode: Callable
    decode: Callable


js_codec = Codec("", j_dumps, j_loads)
mp_codec = Codec("application/x-msgpack", ws_pack, ws_unpack)

CODECS_PARAMS = list(itertools.product((js_codec, mp_codec), (js_codec, mp_codec)))


@pytest.mark.asyncio  # @UndefinedVariable
@pytest.mark.parametrize("send,receive", CODECS_PARAMS)
async def test_get_data(send, receive):
    line = LineData(
        key="new_line",
        line_params=LineParams(colour="orange"),
        x=[5, 6, 7, 8, 9],
        y=[20, 30, 40, 50, 60],
    )

    new_line = PlotMessage(
        plot_id="plot_0", type=MsgType.new_multiline_data, params=[line]
    )

    app = _create_bare_app()

    async with AsyncClient(app=app, base_url="http://test") as ac:
        headers = {}
        if send.mime_type:
            headers["Content-Type"] = send.mime_type
        if receive.mime_type:
            headers["Accept"] = receive.mime_type
        msg = send.encode(new_line)
        response = await ac.post("/push_data", content=msg, headers=headers)
        assert response.status_code == 200
        assert receive.decode(response._content) == "data sent"


@pytest.mark.asyncio
async def test_clear_data_via_message():
    app = _create_bare_app()

    with TestClient(app) as client:
        ps = getattr(app, "_plot_server")

        with client.websocket_connect("/plot/8123f452/plot_0"):
            with client.websocket_connect("/plot/fc8ed0e5/plot_1"):
                async with AsyncClient(app=app, base_url="http://test") as ac:
                    response = await ac.put(
                        "/clear_data/plot_0",
                        params={},
                        headers={"Content-Type": "application/json"},
                    )
                    assert response.status_code == 200
                    assert response.json() == "data cleared"

                    response = await ac.put(
                        "/clear_data/plot_1",
                        params={},
                        headers={"Content-Type": "application/json"},
                    )
                    assert response.status_code == 200
                    assert response.json() == "data cleared"

                plot_state_0 = ps.plot_states["plot_0"]
                assert plot_state_0.new_data_message is None
                assert plot_state_0.new_selections_message is None
                assert plot_state_0.current_data is None
                assert plot_state_0.current_selections is None


@pytest.mark.asyncio
async def test_push_points():
    x = [i for i in range(10)]
    y = [j % 10 for j in x]
    time_id = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    line = LineData(
        key=time_id,
        line_params=LineParams(colour="purple"),
        x=x,
        y=y,
    )
    new_line = PlotMessage(
        plot_id="plot_0", type=MsgType.new_multiline_data, params=[line]
    )
    msg = ws_pack(new_line)
    headers = {
        "Content-Type": "application/x-msgpack",
        "Accept": "application/x-msgpack",
    }
    app = _create_bare_app()

    with TestClient(app) as client:
        with client.websocket_connect("/plot/99a81b01/plot_0"):
            async with AsyncClient(app=app, base_url="http://test") as ac:
                response = await ac.post("/push_data", content=msg, headers=headers)
            assert response.status_code == 200
            assert ws_unpack(response._content) == "data sent"


class TrialA(NumpyModel):
    integers: list[int]
    floats: list[float]
    array: DvDNDArray | None = None


class TrialB(NumpyModel):
    integers: list[int]
    floats: list[float]
    array: DvDNDArray | None = None
    original: TrialA


_nptest_assert_eq = np.testing.assert_equal  # supports dict testing


def nppd_assert_equal(this, other: Any) -> None:
    is_enum = isinstance(this, Enum)
    assert is_enum or isinstance(this, type(other))
    if is_enum:
        assert this.value == other
    elif isinstance(this, dict):
        assert len(this) == len(other)
        for k, v in this.items():
            assert k in other
            nppd_assert_equal(v, other[k])
    elif isinstance(this, (list, tuple)):
        assert len(this) == len(other)
        for t, o in zip(this, other):
            nppd_assert_equal(t, o)
    elif isinstance(this, BaseModel):
        nppd_assert_equal(this.model_dump(), other.model_dump())
    elif isinstance(this, np.ndarray):
        _nptest_assert_eq(this, other)
    elif is_dataclass(this):
        nppd_assert_equal(asdict(this), asdict(other))
    else:
        assert this == other


@pytest.mark.asyncio
@pytest.mark.parametrize("send,receive", CODECS_PARAMS)
async def test_post_test_pydantic(send, receive):
    testa = TrialA(
        integers=[1, 3, 4],
        floats=[-0.5, 1.7, 10.0],
        array=np.array([-1.5, 4]),
    )
    testb = TrialB(
        integers=[0, 2, 3],
        floats=[2.0, 4.2, 12.5],
        array=np.array([-3.75, 10]),
        original=testa,
    )
    app = _create_bare_app()

    @app.post("/test_pydantic")
    @message_unpack
    async def pydantic_test(data: TrialA) -> TrialB:
        result = TrialB(
            integers=[i - 1 for i in data.integers],
            floats=[f + 2.5 for f in data.floats],
            array=data.array * 2.5 if data.array is not None else None,
            original=data,
        )
        return result

    async with AsyncClient(app=app, base_url="http://test") as ac:
        headers = {}
        if send.mime_type:
            headers["Content-Type"] = send.mime_type
        if receive.mime_type:
            headers["Accept"] = receive.mime_type
        msg = send.encode(testa)
        response = await ac.post("/test_pydantic", content=msg, headers=headers)
        assert response.status_code == 200
        nppd_assert_equal(
            TrialB.model_validate(receive.decode(response._content)), testb
        )


if __name__ == "__main__":
    import json

    print(json.dumps(TrialB.model_json_schema(), indent=2))



from math import pi

import numpy as np
import pytest
from davidia.models.selections import (
    AnySelection,
    AxialSelection,
    CircularSectorialSelection,
    CircularSelection,
    EllipticalSelection,
    LinearSelection,
    PolygonalSelection,
    RectangularSelection,
    SelectionBase,
    as_selection,
)
from pydantic.main import BaseModel

parameters = [
    (
        AxialSelection,
        dict(start=(1, 2), dimension=0, length=3.5),
        dict(length=3.5),
        dict(length=4),
    ),
    (
        LinearSelection,
        dict(start=(1, 2), length=3.5, degrees=30),
        dict(degrees=30),
        dict(angle=1),
    ),
    (
        RectangularSelection,
        dict(start=(1, 2), lengths=(3.5, 1.7), angle=pi / 3),
        dict(lengths=(3.5, 1.7)),
        dict(angle=1),
    ),
    (
        PolygonalSelection,
        dict(start=None, points=[(3.5, 1.7), (4.1, -3)]),
        dict(start=(3.5, 1.7)),
        dict(points=[(3.5, -1.7)]),
    ),
    (CircularSelection, dict(start=(1, 2), radius=3), dict(radius=3.0), dict(radius=2)),
    (
        EllipticalSelection,
        dict(start=(1, 2), semi_axes=(2, 3.5)),
        dict(semi_axes=(2, 3.5)),
        dict(semi_axes=(3.2, 8)),
    ),
    (
        CircularSectorialSelection,
        dict(start=(1, 2), radii=(3, 5)),
        dict(radii=(3.0, 5.0)),
        dict(radii=(2, 6)),
    ),
]


def create_from_parameters(param):
    return param[0](**param[1])


@pytest.mark.parametrize("s_cls,args,check,change", parameters)
def test_selection(s_cls, args, check, change):
    ts = s_cls(**args)
    print(repr(ts))
    for f, v in check.items():
        assert np.all(np.isclose(getattr(ts, f), v))
    print(ts.model_dump())
    for f, v in change.items():
        setattr(ts, f, v)
    for f, v in change.items():
        assert np.all(np.isclose(getattr(ts, f), v))
    assert isinstance(as_selection(ts.model_dump()), s_cls)


def test_polygonal():
    with pytest.raises(ValueError):
        PolygonalSelection()
    with pytest.raises(ValueError):
        PolygonalSelection(points=[])
    with pytest.raises(ValueError):
        PolygonalSelection(start=[1])
    with pytest.raises(ValueError):
        PolygonalSelection(start=[1, 2, 3])

    ps = PolygonalSelection(points=[(2, 3)])
    assert all(np.isclose(ps.start, ps.points[0]))
    ps = PolygonalSelection(start=(2, 3))
    assert all(np.isclose(ps.start, ps.points[0]))
    ps = PolygonalSelection(start=(2, 3), points=[])
    assert all(np.isclose(ps.start, ps.points[0]))
    ps = PolygonalSelection(start=(2, 3), points=[(3, 2)])  # warns
    assert all(np.isclose(ps.start, (3, 2)))
    assert ps.closed
    ps = PolygonalSelection(start=(2, 3), points=[(2, 3), (3, 2)], closed=False)
    assert all(np.isclose(ps.start, (2, 3)))
    assert not ps.closed


def test_degrees():
    ls = LinearSelection(start=(1, 2), length=3.5, degrees=30)
    assert np.isclose(ls.angle, pi / 6)
    assert np.isclose(ls.degrees, 30)
    ls = LinearSelection(start=(1, 2), length=3.5, angle=pi / 3)
    assert np.isclose(ls.angle, pi / 3)
    assert np.isclose(ls.degrees, 60)
    ls.set_degrees(45)
    assert np.isclose(ls.angle, pi / 4)
    assert np.isclose(ls.degrees, 45)
    cs = CircularSectorialSelection(start=(1, 2), radii=(3.5, 5))
    assert all(np.isclose(cs.angles, (0, 2 * pi)))
    assert all(np.isclose(cs.degrees, (0, 360)))
    cs = CircularSectorialSelection(start=(1, 2), radii=(3.5, 5), degrees=(30, 120))
    assert all(np.isclose(cs.angles, (pi / 6, 2 * pi / 3)))
    assert all(np.isclose(cs.degrees, (30, 120)))
    cs.set_degrees((45, 180))
    assert all(np.isclose(cs.angles, (pi / 4, pi)))
    assert all(np.isclose(cs.degrees, (45, 180)))


class MySBModel(BaseModel):
    base: SelectionBase
    any_sel: AnySelection


def test_generic_model():
    b = create_from_parameters(parameters[1])
    c = create_from_parameters(parameters[2])
    x = MySBModel(base=b, any_sel=c)
    dx = x.model_dump()
    print(dx)
    assert "start" in dx["base"]
    assert "angle" not in dx["base"]
    assert "length" not in dx["base"]
    assert "start" in dx["any_sel"]
    assert "angle" in dx["any_sel"]
    assert "lengths" in dx["any_sel"]


if __name__ == "__main__":
    test_generic_model()
    test_polygonal()
    p = create_from_parameters(parameters[3])
